#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Markus Löning"]
__all__ = ["Tabularizer"]

import pandas as pd
from sklearn.utils.validation import check_array

from sktime.transformations.base import _PanelToTabularTransformer
from sktime.utils.data_processing import from_2d_array_to_nested
from sktime.utils.data_processing import from_3d_numpy_to_2d_array
from sktime.utils.data_processing import from_nested_to_2d_array
from sktime.utils.validation.panel import check_X


class Tabularizer(_PanelToTabularTransformer):
    """
    A transformer that turns time series/panel data into tabular data.

    This estimator converts nested pandas dataframe containing
    time-series/panel data with numpy arrays or pandas Series in
    dataframe cells into a tabular pandas dataframe with only primitives in
    cells. This is useful for transforming
    time-series/panel data into a format that is accepted by standard
    validation learning algorithms (as in sklearn).
    """

    def transform(self, X, y=None):
        """Transform nested pandas dataframe into tabular dataframe.

        Parameters
        ----------
        X : pandas DataFrame
            Nested dataframe with pandas series or numpy arrays in cells.
        y : array-like, optional (default=None)

        Returns
        -------
        Xt : pandas DataFrame
            Transformed dataframe with only primitives in cells.
        """
        self.check_is_fitted()
        X = check_X(X)
        if isinstance(X, pd.DataFrame):
            return from_nested_to_2d_array(X)
        else:
            return from_3d_numpy_to_2d_array(X)

    def inverse_transform(self, X, y=None):
        """Transform tabular pandas dataframe into nested dataframe.

        Parameters
        ----------
        X : pandas DataFrame
            Tabular dataframe with primitives in cells.
        y : array-like, optional (default=None)

        Returns
        -------
        Xt : pandas DataFrame
            Transformed dataframe with series in cells.
        """
        self.check_is_fitted()
        # We expect a tabular pd.DataFrame or np.array here, hence we use
        # scikit-learn's input validation function.
        X = check_array(X)
        return from_2d_array_to_nested(X)



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Markus Löning"]
__all__ = ["ESTIMATOR_TEST_PARAMS", "EXCLUDE_ESTIMATORS", "EXCLUDED_TESTS"]

import numpy as np
from hcrystalball.wrappers import HoltSmoothingWrapper
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import FunctionTransformer
from sklearn.preprocessing import StandardScaler
from sktime.forecasting.fbprophet import Prophet
from sktime.base import BaseEstimator
from sktime.classification.base import BaseClassifier
from sktime.classification.compose import ColumnEnsembleClassifier
from sktime.classification.compose import TimeSeriesForestClassifier
from sktime.classification.dictionary_based import TemporalDictionaryEnsemble
from sktime.classification.frequency_based import RandomIntervalSpectralForest
from sktime.classification.interval_based import TimeSeriesForest
from sktime.classification.shapelet_based import ShapeletTransformClassifier
from sktime.forecasting.arima import AutoARIMA
from sktime.forecasting.base import BaseForecaster
from sktime.forecasting.bats import BATS
from sktime.forecasting.compose import DirectRegressionForecaster
from sktime.forecasting.compose import DirectTimeSeriesRegressionForecaster
from sktime.forecasting.compose import EnsembleForecaster
from sktime.forecasting.compose import RecursiveRegressionForecaster
from sktime.forecasting.compose import RecursiveTimeSeriesRegressionForecaster
from sktime.forecasting.compose import StackingForecaster
from sktime.forecasting.compose import TransformedTargetForecaster
from sktime.forecasting.exp_smoothing import ExponentialSmoothing
from sktime.forecasting.hcrystalball import HCrystalBallForecaster
from sktime.forecasting.model_selection import ForecastingGridSearchCV
from sktime.forecasting.model_selection import SingleWindowSplitter
from sktime.forecasting.naive import NaiveForecaster
from sktime.forecasting.online_learning import OnlineEnsembleForecaster
from sktime.forecasting.tbats import TBATS
from sktime.forecasting.theta import ThetaForecaster
from sktime.performance_metrics.forecasting import sMAPE
from sktime.regression.base import BaseRegressor
from sktime.regression.compose import TimeSeriesForestRegressor
from sktime.series_as_features.compose import FeatureUnion
from sktime.transformations.base import BaseTransformer
from sktime.transformations.base import _PanelToPanelTransformer
from sktime.transformations.base import _PanelToTabularTransformer
from sktime.transformations.base import _SeriesToPrimitivesTransformer
from sktime.transformations.base import _SeriesToSeriesTransformer
from sktime.transformations.panel.compose import ColumnTransformer
from sktime.transformations.panel.compose import (
    SeriesToPrimitivesRowTransformer,
)
from sktime.transformations.panel.compose import SeriesToSeriesRowTransformer
from sktime.transformations.panel.dictionary_based import SFA
from sktime.transformations.panel.interpolate import TSInterpolator
from sktime.transformations.panel.reduce import Tabularizer
from sktime.transformations.panel.shapelets import ContractedShapeletTransform
from sktime.transformations.panel.shapelets import ShapeletTransform
from sktime.transformations.panel.summarize import FittedParamExtractor
from sktime.transformations.panel.tsfresh import TSFreshFeatureExtractor
from sktime.transformations.panel.tsfresh import (
    TSFreshRelevantFeatureExtractor,
)
from sktime.transformations.series.acf import AutoCorrelationTransformer
from sktime.transformations.series.acf import PartialAutoCorrelationTransformer
from sktime.transformations.series.adapt import TabularToSeriesAdaptor
from sktime.transformations.series.detrend import Detrender

# The following estimators currently do not pass all unit tests or fail some of them
# and are excluded until fixed.
EXCLUDE_ESTIMATORS = [
    "ElasticEnsemble",
    "KNeighborsTimeSeriesClassifier",
    "ProximityForest",
    "ProximityStump",
    "ProximityTree",
]

EXCLUDED_TESTS = {
    "ShapeletTransformClassifier": ["check_fit_idempotent"],
    "ContractedShapeletTransform": ["check_fit_idempotent"],
}

# We here configure estimators for basic unit testing, including setting of
# required hyper-parameters and setting of hyper-parameters for faster training.
SERIES_TO_SERIES_TRANSFORMER = StandardScaler()
SERIES_TO_PRIMITIVES_TRANSFORMER = FunctionTransformer(
    np.mean, kw_args={"axis": 0}, check_inverse=False
)
TRANSFORMERS = [
    (
        "transformer1",
        SeriesToSeriesRowTransformer(
            SERIES_TO_SERIES_TRANSFORMER, check_transformer=False
        ),
    ),
    (
        "transformer2",
        SeriesToSeriesRowTransformer(
            SERIES_TO_SERIES_TRANSFORMER, check_transformer=False
        ),
    ),
]
REGRESSOR = LinearRegression()
TIME_SERIES_CLASSIFIER = TimeSeriesForest(n_estimators=3)
TIME_SERIES_CLASSIFIERS = [
    ("tsf1", TIME_SERIES_CLASSIFIER),
    ("tsf2", TIME_SERIES_CLASSIFIER),
]
FORECASTER = ExponentialSmoothing()
FORECASTERS = [("ses1", FORECASTER), ("ses2", FORECASTER)]
STEPS = [
    ("transformer", Detrender(ThetaForecaster())),
    ("forecaster", NaiveForecaster()),
]
ESTIMATOR_TEST_PARAMS = {
    OnlineEnsembleForecaster: {"forecasters": FORECASTERS},
    FeatureUnion: {"transformer_list": TRANSFORMERS},
    DirectRegressionForecaster: {"regressor": REGRESSOR},
    RecursiveRegressionForecaster: {"regressor": REGRESSOR},
    DirectTimeSeriesRegressionForecaster: {
        "regressor": make_pipeline(Tabularizer(), REGRESSOR)
    },
    RecursiveTimeSeriesRegressionForecaster: {
        "regressor": make_pipeline(Tabularizer(), REGRESSOR)
    },
    TransformedTargetForecaster: {"steps": STEPS},
    EnsembleForecaster: {"forecasters": FORECASTERS},
    StackingForecaster: {"forecasters": FORECASTERS, "final_regressor": REGRESSOR},
    Detrender: {"forecaster": FORECASTER},
    ForecastingGridSearchCV: {
        "forecaster": NaiveForecaster(strategy="mean"),
        "cv": SingleWindowSplitter(fh=1),
        "param_grid": {"window_length": [2, 5]},
        "scoring": sMAPE(),
    },
    TabularToSeriesAdaptor: {"transformer": StandardScaler()},
    ColumnEnsembleClassifier: {
        "estimators": [
            (name, estimator, 0) for (name, estimator) in TIME_SERIES_CLASSIFIERS
        ]
    },
    FittedParamExtractor: {
        "forecaster": FORECASTER,
        "param_names": ["initial_level"],
    },
    SeriesToPrimitivesRowTransformer: {
        "transformer": SERIES_TO_PRIMITIVES_TRANSFORMER,
        "check_transformer": False,
    },
    SeriesToSeriesRowTransformer: {
        "transformer": SERIES_TO_SERIES_TRANSFORMER,
        "check_transformer": False,
    },
    ColumnTransformer: {
        "transformers": [(name, estimator, [0]) for name, estimator in TRANSFORMERS]
    },
    AutoARIMA: {
        "d": 0,
        "suppress_warnings": True,
        "max_p": 2,
        "max_q": 2,
        "seasonal": False,
    },
    ShapeletTransformClassifier: {"n_estimators": 3, "time_contract_in_mins": 0.125},
    ContractedShapeletTransform: {"time_contract_in_mins": 0.125},
    ShapeletTransform: {
        "max_shapelets_to_store_per_class": 1,
        "min_shapelet_length": 3,
        "max_shapelet_length": 4,
    },
    TSFreshFeatureExtractor: {"disable_progressbar": True, "show_warnings": False},
    TSFreshRelevantFeatureExtractor: {
        "disable_progressbar": True,
        "show_warnings": False,
        "fdr_level": 0.01,
    },
    TSInterpolator: {"length": 10},
    RandomIntervalSpectralForest: {"n_estimators": 3, "acf_lag": 10, "min_interval": 5},
    SFA: {"return_pandas_data_series": True},
    TemporalDictionaryEnsemble: {
        "n_parameter_samples": 50,
        "max_ensemble_size": 10,
        "randomly_selected_params": 40,
    },
    TimeSeriesForest: {"n_estimators": 3},
    TimeSeriesForestClassifier: {"n_estimators": 3},
    TimeSeriesForestRegressor: {"n_estimators": 3},
    HCrystalBallForecaster: {"model": HoltSmoothingWrapper()},
    BATS: {
        "use_box_cox": False,
        "use_trend": False,
        "use_damped_trend": False,
        "sp": [],
        "use_arma_errors": False,
        "n_jobs": 1,
    },
    TBATS: {
        "use_box_cox": False,
        "use_trend": False,
        "use_damped_trend": False,
        "sp": [],
        "use_arma_errors": False,
        "n_jobs": 1,
    },
    Prophet: {
        "n_changepoints": 0,
        "yearly_seasonality": False,
        "weekly_seasonality": False,
        "daily_seasonality": False,
        "uncertainty_samples": 1000,
        "verbose": False,
    },
    PartialAutoCorrelationTransformer: {"n_lags": 1},
    AutoCorrelationTransformer: {"n_lags": 1},
}

# These methods should not change the state of the estimator, that is, they should
# not change fitted parameters or hyper-parameters. They are also the methods that
# "apply" the fitted estimator to data and useful for checking results.
NON_STATE_CHANGING_METHODS = (
    "predict",
    "predict_proba",
    "decision_function",
    "transform",
    "inverse_transform",
)

# We use estimator tags in addition to class hierarchies to further distinguish
# estimators into different categories. This is useful for defining and running
# common tests for estimators with the same tags.
VALID_ESTIMATOR_TAGS = (
    "fit-in-transform",  # fitted in transform or non-fittable
    "univariate-only",
    "transform-returns-same-time-index",
)

# The following gives a list of valid estimator base classes.
VALID_TRANSFORMER_TYPES = (
    _SeriesToPrimitivesTransformer,
    _SeriesToSeriesTransformer,
    _PanelToTabularTransformer,
    _PanelToPanelTransformer,
)
VALID_ESTIMATOR_BASE_TYPES = (
    BaseClassifier,
    BaseRegressor,
    BaseForecaster,
    BaseTransformer,
)
VALID_ESTIMATOR_TYPES = (
    BaseEstimator,
    *VALID_ESTIMATOR_BASE_TYPES,
    *VALID_TRANSFORMER_TYPES,
)

VALID_ESTIMATOR_BASE_TYPE_LOOKUP = {
    "classifier": BaseClassifier,
    "regressor": BaseRegressor,
    "forecaster": BaseForecaster,
    "transformer": BaseTransformer,
}



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import pytest

from sktime.transformations.panel.summarize import PlateauFinder


@pytest.mark.parametrize("value", [np.nan, -10, 10, -0.5, 0.5])
def test_PlateauFinder(value):
    # generate test data
    value = np.nan
    X = pd.DataFrame(
        pd.Series(
            [
                pd.Series([value, 1, 2, 3, value, 2, 2, 3]),  # nan at start
                pd.Series([value, value, 3, 3, value, 2, 2, 3]),
                pd.Series([0, value, value, value, value, value, 2, value]),
                # nan at end
                pd.Series([0, value, value, 3, 4, 5, value, value]),
                pd.Series([2, value, value, value, 2, value, 3, 1]),
                pd.Series([0, value, value, 3, value, value, 2, 0]),
            ]
        )
    )
    n_samples = X.shape[0]

    t = PlateauFinder(value=value, min_length=2)
    Xt = t.fit_transform(X)

    actual_starts = Xt.iloc[:, 0]
    actual_lengths = Xt.iloc[:, 1]

    expected_starts = [
        np.array([]),
        np.array([0]),
        np.array([1]),
        np.array([1, 6]),
        np.array([1]),
        np.array([1, 4]),
    ]
    expected_lengths = [
        np.array([]),
        np.array([2]),
        np.array([5]),
        np.array([2, 2]),
        np.array([3]),
        np.array([2, 2]),
    ]

    # compare results
    for i in range(n_samples):
        np.testing.assert_array_equal(actual_starts[i], expected_starts[i])
        np.testing.assert_array_equal(actual_lengths[i], expected_lengths[i])



import numpy as np
import os
from sklearn.metrics.scorer import accuracy_score
from sktime.contrib.experiments import write_results_to_uea_format


class ElasticEnsemblePostProcess:
    __author__ = "Jason Lines"

    def __init__(self, results_path, dataset_name, distance_measures='all', resample_id=0, alpha=1):
        """

        Args:
            results_path: String - path to folder storing the results to be read into the ensemble
            dataset_name: String - the name of the dataset that this ensemble will post process results for
            distance_measures: 'all' or list of Strings - default 'all'. 'all' sets classifier to use all default constituent classifiers,
                                else a list is provided of classifiers to include. Note these names must match the names of
                                the subdirs in the folder located at results_parh
            resample_id: default = 0 - to identify the deterministic seed used for resampling experiments.
                         A resampled_id of 0 demonstrates default train/test split were used to create results
            alpha: float/double - default=1.0. Used to exponentiate the confidence of constituent classifiers when making test predictions
        """
        self.results_path = results_path
        self.dataset_name = dataset_name
        self.resample_id = resample_id
        if distance_measures == 'all':
            self.distance_measures=['dtw','ddtw','wdtw','wddtw','lcss','erp','msm']
        else:
            self.distance_measures = distance_measures
        self.alpha=alpha

        # load in train information
        self.train_accs_by_classifier = np.zeros(len(self.distance_measures))
        self.train_dists_by_classifier = []
        self.test_dists_by_classifier = []
        self.ee_train_dists = None
        self.ee_test_dists = None

        self.actual_train_class_vals = None
        self.actual_test_class_vals = None
        self.classes_ = None

        num_classes = None
        num_ins = None
        class_vals = None
        # load train

        for c_id in range(len(self.distance_measures)):
            file_path = self.results_path+self.distance_measures[c_id]+"/Predictions/"+self.dataset_name+"/trainFold"+str(self.resample_id)+".csv"
            with open(file_path, 'r') as f:
                lines = f.readlines()
                third_line = lines[2].split(",")
                self.train_accs_by_classifier[c_id] = float(third_line[0].strip())
                this_class_vals = third_line[-1].strip().replace("[","").replace("]","").split(" ")
                this_num_classes = len(this_class_vals)
                this_num_ins = len(lines)-3
                if class_vals is None:
                    class_vals = this_class_vals
                    self.classes_ = np.array(this_class_vals)
                    num_classes = this_num_classes
                    num_ins = this_num_ins

                elif this_class_vals != class_vals:
                    raise ValueError("Class value mismatch when loading train file for "
                                     + str(self.distance_measures[c_id])+" and "+self.dataset_name)
                elif this_num_ins != num_ins:
                    raise ValueError("Inconsistent number of predictions in constituent training files: first spotted "
                                     "in train file for "+str(self.distance_measures[c_id])+" and "+self.dataset_name)

                this_dists = np.empty((num_ins,num_classes))
                this_actual_train_class_vals = []
                for i in range(num_ins):
                    split_line = lines[i+3].strip().split(",")
                    this_actual_train_class_vals.append(split_line[0].strip())
                    for c in range(num_classes):
                        this_dists[i][c] = np.power(float(split_line[c+3]),self.alpha)*self.train_accs_by_classifier[c_id]
                if self.actual_train_class_vals is None:
                    self.actual_train_class_vals = this_actual_train_class_vals
                elif self.actual_train_class_vals != this_actual_train_class_vals:
                    raise ValueError("Class values in files no not match for train - first spotted for "+str(self.distance_measures[c_id]))

            if self.ee_train_dists is None:
                self.ee_train_dists = this_dists
            else:
                self.ee_train_dists = np.add(self.ee_train_dists,this_dists)
            self.train_dists_by_classifier.append(this_dists)
        self.ee_train_dists = np.divide(self.ee_train_dists,sum(self.train_accs_by_classifier))

        # load test
        num_test_ins = None
        for c_id in range(len(self.distance_measures)):
            file_path = self.results_path + self.distance_measures[c_id] + "/Predictions/" + self.dataset_name + "/testFold" + str(self.resample_id) + ".csv"
            with open(file_path, 'r') as f:
                lines = f.readlines()
                third_line = lines[2].split(",")
                this_class_vals = third_line[-1].strip().replace("[", "").replace("]", "").split(" ")
                this_num_ins = len(lines) - 3
                if this_class_vals != class_vals:
                    raise ValueError("Class value mismatch when loading test file for "
                                     + str(self.distance_measures[c_id])+" and "+self.dataset_name)
                if num_test_ins is None:
                    num_test_ins = this_num_ins
                elif num_test_ins != this_num_ins:
                    raise ValueError("Inconsistent number of predictions in constituent test files: first spotted "
                                     "in train file for " + str(self.distance_measures[c_id]) + " and " + self.dataset_name)
                this_dists = np.empty((num_ins, num_classes))
                this_actual_test_class_vals = []
                for i in range(num_ins):
                    split_line = lines[i + 3].strip().split(",")
                    this_actual_test_class_vals.append(split_line[0].strip())
                    for c in range(num_classes):
                        this_dists[i][c] = np.power(float(split_line[c + 3]), self.alpha) * self.train_accs_by_classifier[c_id]
                if self.actual_test_class_vals is None:
                    self.actual_test_class_vals = this_actual_test_class_vals
                elif self.actual_test_class_vals != this_actual_test_class_vals:
                    raise ValueError("Class values in files no not match for test - first spotted for "+str(self.distance_measures[c_id]))

            if self.ee_test_dists is None:
                self.ee_test_dists = this_dists
            else:
                self.ee_test_dists = np.add(self.ee_test_dists, this_dists)
            self.test_dists_by_classifier.append(this_dists)

        self.ee_test_dists = np.divide(self.ee_test_dists, sum(self.train_accs_by_classifier))

    def write_files(self, output_results_path, output_classifier_name="EE", write_train=True, write_test=True, overwrite=False):
        """

        Args:
            output_results_path: String - path to where output results will be written
            output_classifier_name: String - the name of the composite ensemble classifier in the output files
            write_train: boolean - true will write train files for the ensemble, false will skip training files
            write_test: boolean - true will write test files for the ensemble, false will skip test files
            overwrite: boolean - if true, any existing train/test files will be over-written. False prevents file overwriting


        """
        if write_train is False and write_test is False:
            print("Train and test writing both set to false - method will terminate without doing anything")
            return

        if not overwrite:
            if write_train:
                full_path = str(output_results_path) + "/" + str(output_classifier_name) + "/Predictions/" + str(self.dataset_name) + "/trainFold" + str(self.resample_id) + ".csv"
                if os.path.exists(full_path):
                    print(full_path + " already exists and overwrite set to false, not writing Train", Warning)
                    write_train = False

            if write_test is True:
                full_path = str(output_results_path) + "/" + str(output_classifier_name) + "/Predictions/" + str(self.dataset_name) + "/testFold" + str(self.resample_id) + ".csv"
                if os.path.exists(full_path):
                    print(full_path + " already exists and overwrite set to false, not writing Test")
                    write_test = False

        if write_train is False and write_test is False:
            print("Train and test files both already exist and overwrite set to false - method will terminate without doing anything")
            return

        """
        file_format = None
        if os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name+ '_TRAIN.ts') and os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name+ '_TEST.ts'):
            train_x, train_y = loader.load_from_tsfile_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TRAIN.ts')
            test_x, test_y = loader.load_from_tsfile_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TEST.ts')
        elif os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name+ '_TRAIN.arff') and os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name+ '_TEST.arff'):
            train_x, train_y = loader.load_from_arff_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TRAIN.ts')
            test_x, test_y = loader.load_from_arff_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TEST.ts')
        elif os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name + '_TRAIN.tsv') and os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name + '_TEST.tsv'):
            train_x, train_y = loader.load_from_ucr_tsv_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TRAIN.ts')
            test_x, test_y = loader.load_from_ucr_tsv_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TEST.ts')
        else:
            raise ValueError("No dataset found for "+self.dataset_name)
        """
        if write_train:
            train_probs = self.ee_train_dists
            train_preds = self.classes_[np.argmax(train_probs, axis=1)]
            acc = accuracy_score(self.actual_train_class_vals, train_preds)
            second = str(self.distance_measures)
            third = str(acc) + ",NA,NA,-1,-1," + str(len(self.classes_)) + "," + str(self.classes_)
            write_results_to_uea_format(second_line=second, third_line=third, output_path=output_results_path, classifier_name=output_classifier_name, resample_seed=self.resample_id,
                                        predicted_class_vals=train_preds, actual_probas=train_probs, dataset_name=self.dataset_name, actual_class_vals=self.actual_train_class_vals, split='TRAIN')

        if write_test:
            test_probs = self.ee_test_dists
            test_preds = self.classes_[np.argmax(test_probs, axis=1)]
            acc = accuracy_score(self.actual_test_class_vals, test_preds)
            second = str(self.distance_measures)
            third = str(acc) + ",NA,NA,-1,-1," + str(len(self.classes_)) + "," + str(self.classes_)
            write_results_to_uea_format(second_line=second, third_line=third, output_path=output_results_path, classifier_name=output_classifier_name, resample_seed=self.resample_id,
                                        predicted_class_vals=test_preds, actual_probas=test_probs, dataset_name=self.dataset_name, actual_class_vals=self.actual_test_class_vals, split='TEST')




__all__ = [
    "MrSEQLClassifier",
    "ShapeletTransformClassifier"
]

from sktime.classification.shapelet_based.mrseql.mrseql import \
    MrSEQLClassifier
from sktime.classification.shapelet_based._stc import \
    ShapeletTransformClassifier



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Tomasz Chodakowski"]

import pytest

from sktime.performance_metrics.forecasting import mape_loss
from sktime.performance_metrics.forecasting import smape_loss

from sktime.performance_metrics.tests._config import TEST_YS, TEST_YS_ZERO


@pytest.mark.parametrize("test_y", TEST_YS_ZERO)
def test_mape_loss_near_zero(test_y):
    # mape should be large when predicitons are close to zero
    y_test = test_y
    y_pred = y_test + 0.001
    assert mape_loss(y_test, y_pred) > 1e6


@pytest.mark.parametrize("test_y", TEST_YS)
def test_mape_loss(test_y):
    y_test = test_y
    y_pred = y_test
    assert mape_loss(y_test, y_pred) == pytest.approx(0.0)

    y_pred = y_test * 1.1
    assert mape_loss(y_test, y_pred) == pytest.approx(0.1)

    y_pred = y_test * 0.9
    assert mape_loss(y_test, y_pred) == pytest.approx(0.1)

    y_pred = y_test * 1.000001
    assert mape_loss(y_test, y_pred) == pytest.approx(0.000001)

    y_pred = y_test * 2
    assert mape_loss(y_test, y_pred) == pytest.approx(1.0)

    y_pred = y_test * 100
    assert mape_loss(y_test, y_pred) == pytest.approx(99.0)


@pytest.mark.parametrize("test_y", TEST_YS)
def test_smape_loss(test_y):
    y_test = test_y
    y_pred = y_test
    assert smape_loss(y_test, y_pred) == pytest.approx(0.0)
    assert smape_loss(y_pred, y_test) == pytest.approx(0.0)

    y_pred = y_test * 1.1
    assert smape_loss(y_test, y_pred) == pytest.approx(0.095238095238)
    assert smape_loss(y_pred, y_test) == pytest.approx(0.095238095238)

    y_pred = y_test * 1.000001
    assert smape_loss(y_test, y_pred) == pytest.approx(0.000001)
    assert smape_loss(y_pred, y_test) == pytest.approx(0.000001)

    y_pred = y_test * 2.0
    assert smape_loss(y_test, y_pred) == pytest.approx(0.6666666)
    assert smape_loss(y_pred, y_test) == pytest.approx(0.6666666)

    y_pred = y_test * 100
    assert smape_loss(y_test, y_pred) == smape_loss(y_pred, y_test)



# -*- coding: utf-8 -*-
from sktime.utils.data_io import load_from_tsfile_to_dataframe as load_ts


def test_loading():
    # test multivariate
    # Test univariate
    data_dir = "E:/tsc_ts/"
    dataset = "Gunpoint"
    trainX, trainY = load_ts(data_dir + dataset + "/" + dataset + "_TRAIN.ts")
    testX, testY = load_ts(data_dir + dataset + "/" + dataset + "_TEST.ts")
    print("Loaded " + dataset + " in position " + str(i))


if __name__ == "__main__":
    """
    Example simple usage, with arguments input via script or hard coded for testing
    """



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from joblib import Parallel
from joblib import delayed
from scipy import sparse
from sklearn.pipeline import FeatureUnion as _FeatureUnion
from sklearn.pipeline import _fit_transform_one
from sklearn.pipeline import _transform_one

from sktime.transformations.base import _PanelToPanelTransformer

__all__ = ["FeatureUnion"]
__author__ = ["Markus Löning"]


class FeatureUnion(_FeatureUnion, _PanelToPanelTransformer):
    """Concatenates results of multiple transformer objects.
    This estimator applies a list of transformer objects in parallel to the
    input data, then concatenates the results. This is useful to combine
    several feature extraction mechanisms into a single transformer.
    Parameters of the transformations may be set using its name and the
    parameter
    name separated by a '__'. A transformer may be replaced entirely by
    setting the parameter with its name to another transformer,
    or removed by setting to 'drop' or ``None``.
    Parameters
    ----------
    transformer_list : list of (string, transformer) tuples
        List of transformer objects to be applied to the data. The first
        half of each tuple is the name of the transformer.
    n_jobs : int or None, optional (default=None)
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
        context.
        ``-1`` means using all processors.
    transformer_weights : dict, optional
        Multiplicative weights for features per transformer.
        Keys are transformer names, values the weights.
    """

    _required_parameters = ["transformer_list"]

    def __init__(
        self,
        transformer_list,
        n_jobs=None,
        transformer_weights=None,
        preserve_dataframe=True,
    ):
        self.preserve_dataframe = preserve_dataframe
        super(FeatureUnion, self).__init__(
            transformer_list, n_jobs=n_jobs, transformer_weights=transformer_weights
        )

        # We need to add is-fitted state when inheriting from scikit-learn
        self._is_fitted = False

    def fit_transform(self, X, y=None, **fit_params):
        """Fit all transformations, transform the data and concatenate results.
        Parameters
        ----------
        X : pandas DataFrame
            Input data to be transformed.
        y : pandas Series, shape (n_samples, ...), optional
            Targets for supervised learning.

        Returns
        -------
        Xt : pandas DataFrame
            hstack of results of transformations. sum_n_components is the
            sum of n_components (output dimension) over transformations.
        """
        self._validate_transformers()
        result = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_transform_one)(trans, X, y, weight, **fit_params)
            for name, trans, weight in self._iter()
        )

        if not result:
            # All transformations are None
            return np.zeros((X.shape[0], 0))

        Xs, transformers = zip(*result)
        self._update_transformer_list(transformers)

        Xs = self._hstack(list(Xs))
        self._is_fitted = True
        return Xs

    def fit(self, X, y=None, **fit_params):
        super(FeatureUnion, self).fit(X, y, **fit_params)
        self._is_fitted = True
        return self

    def transform(self, X):
        """Transform X separately by each transformer, concatenate results.
        Parameters
        ----------
        X : pandas DataFrame
            Input data to be transformed.
        Returns
        -------
        Xt : pandas DataFrame
            hstack of results of transformations. sum_n_components is the
            sum of n_components (output dimension) over transformations.
        """
        self.check_is_fitted()
        Xs = Parallel(n_jobs=self.n_jobs)(
            delayed(_transform_one)(trans, X, None, weight)
            for name, trans, weight in self._iter()
        )

        if not Xs:
            # All transformations are None
            return np.zeros((X.shape[0], 0))

        else:
            return self._hstack(list(Xs))

    def _hstack(self, Xs):
        """
        Stacks X horizontally.
        Supports input types (X): list of
            numpy arrays, sparse arrays and DataFrames
        """

        if any(sparse.issparse(f) for f in Xs):
            Xs = sparse.hstack(Xs).tocsr()

        types = set(type(X) for X in Xs)
        if self.preserve_dataframe and (pd.Series in types or pd.DataFrame in types):
            return pd.concat(Xs, axis=1)

        else:
            return np.hstack(Xs)



# -*- coding: utf-8 -*-
import os
import tempfile
import pytest
import numpy as np
import pandas as pd
from sktime.utils.data_io import TsFileParseException
from sktime.utils.data_io import load_from_tsfile_to_dataframe
from sktime.utils.data_io import load_from_long_to_dataframe
from sktime.utils.data_io import LongFormatDataParseException
from sktime.utils.data_io import generate_example_long_table


def test_load_from_tsfile_to_dataframe():
    """Test the load_from_tsfile_to_dataframe() function."""

    # Test that an empty file is classed an invalid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = ""

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file and assert that it is invalid

            np.testing.assert_raises(
                TsFileParseException, load_from_tsfile_to_dataframe, path
            )

    finally:
        os.remove(path)

    # Test that a file with an incomplete set of metadata is invalid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps " "true\n@univariate true\n"
            )

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file and assert that it is invalid

            np.testing.assert_raises(
                TsFileParseException, load_from_tsfile_to_dataframe, path
            )

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata but no data is invalid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel false\n@data"
            )

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file and assert that it is invalid

            np.testing.assert_raises(
                TsFileParseException, load_from_tsfile_to_dataframe, path
            )

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and no data but
    # invalid metadata values is invalid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName\n@timeStamps\n@univariate "
                "true\n@classLabel false\n@data"
            )

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file and assert that it is invalid

            np.testing.assert_raises(
                TsFileParseException, load_from_tsfile_to_dataframe, path
            )

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and a single
    # case/dimension parses correctly

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel "
                "false\n@data\n"
            )
            file_contents += "(0, 1), (1, 2)"

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file

            df = load_from_tsfile_to_dataframe(path)

            # Test the DataFrame returned accurately reflects the data in
            # the file

            np.testing.assert_equal(len(df), 1)
            np.testing.assert_equal(len(df.columns), 1)

            series = df["dim_0"]
            np.testing.assert_equal(len(series), 1)

            series = df["dim_0"][0]
            np.testing.assert_equal(series[0], 1.0)
            np.testing.assert_equal(series[1], 2.0)

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and 2 cases with 3
    # dimensions parses correctly

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel "
                "false\n@data\n"
            )
            file_contents += "(0, 1), (1, 2):(0, 3), (1, 4):(0, 5), (1, 6)\n"
            file_contents += "(0, 11), (1, 12):(0, 13), (1,14):(0, 15), (1, 16)     \n"

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file

            df = load_from_tsfile_to_dataframe(path)

            # Test the DataFrame returned accurately reflects the data in
            # the file

            np.testing.assert_equal(len(df), 2)
            np.testing.assert_equal(len(df.columns), 3)

            series = df["dim_0"]
            np.testing.assert_equal(len(series), 2)

            series = df["dim_0"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 1.0)
            np.testing.assert_equal(series[1], 2.0)

            series = df["dim_0"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 11.0)
            np.testing.assert_equal(series[1], 12.0)

            series = df["dim_1"]
            np.testing.assert_equal(len(series), 2)

            series = df["dim_1"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 3.0)
            np.testing.assert_equal(series[1], 4.0)

            series = df["dim_1"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 13.0)
            np.testing.assert_equal(series[1], 14.0)

            series = df["dim_2"]
            np.testing.assert_equal(len(series), 2)

            series = df["dim_2"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 5.0)
            np.testing.assert_equal(series[1], 6.0)

            series = df["dim_2"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 15.0)
            np.testing.assert_equal(series[1], 16.0)

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and time-series of
    # different length parses correctly

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel "
                "false\n@data\n"
            )
            file_contents += "(0, 1), (1, 2):(0, 3):(0, 5), (1, 6)\n"
            file_contents += "(0, 11), (1, 12):(0, 13), (1,14):(0, 15)\n"

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file

            df = load_from_tsfile_to_dataframe(path)

            # Test the DataFrame returned accurately reflects the data in
            # the file

            np.testing.assert_equal(len(df), 2)
            np.testing.assert_equal(len(df.columns), 3)

            series = df["dim_0"]
            np.testing.assert_equal(len(series), 2)

            series = df["dim_0"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 1.0)
            np.testing.assert_equal(series[1], 2.0)

            series = df["dim_0"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 11.0)
            np.testing.assert_equal(series[1], 12.0)

            series = df["dim_1"]
            np.testing.assert_equal(len(series), 2)

            series = df["dim_1"][0]
            np.testing.assert_equal(len(series), 1)
            np.testing.assert_equal(series[0], 3.0)

            series = df["dim_1"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 13.0)
            np.testing.assert_equal(series[1], 14.0)

            series = df["dim_2"]
            np.testing.assert_equal(len(series), 2)

            series = df["dim_2"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 5.0)
            np.testing.assert_equal(series[1], 6.0)

            series = df["dim_2"][1]
            np.testing.assert_equal(len(series), 1)
            np.testing.assert_equal(series[0], 15.0)

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and data but an
    # inconsistent number of dimensions across cases is classed as invalid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel "
                "false\n@data\n"
            )
            file_contents += "(0, 1), (1, 2):(0, 3), (1, 4):(0, 5), (1, 6)\n"
            file_contents += "(0, 11), (1, 12):(0, 13), (1,14)    \n"

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file and assert that it is invalid

            np.testing.assert_raises(
                TsFileParseException, load_from_tsfile_to_dataframe, path
            )

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and data but missing
    # values after a tuple is classed as invalid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel "
                "false\n@data\n"
            )
            file_contents += "(0, 1), (1, 2):(0, 3), (1, 4):(0, 5),\n"

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file and assert that it is invalid

            np.testing.assert_raises(
                TsFileParseException, load_from_tsfile_to_dataframe, path
            )

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and data and some
    # empty dimensions is classed as valid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel "
                "false\n@data\n"
            )
            file_contents += "(0, 1), (1, 2):     :(0, 5), (1, 6)\n"
            file_contents += "(0, 11), (1, 12):(0, 13), (1,14)    :       \n"
            file_contents += (
                "(0, 21), (1, 22):(0, 23), (1,24)    :   (0,25), (1, 26)    \n"
            )

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file

            df = load_from_tsfile_to_dataframe(path)

            # Test the DataFrame returned accurately reflects the data in
            # the file

            np.testing.assert_equal(len(df), 3)
            np.testing.assert_equal(len(df.columns), 3)

            series = df["dim_0"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_0"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 1.0)
            np.testing.assert_equal(series[1], 2.0)

            series = df["dim_0"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 11.0)
            np.testing.assert_equal(series[1], 12.0)

            series = df["dim_0"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 21.0)
            np.testing.assert_equal(series[1], 22.0)

            series = df["dim_1"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_1"][0]
            np.testing.assert_equal(len(series), 0)

            series = df["dim_1"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 13.0)
            np.testing.assert_equal(series[1], 14.0)

            series = df["dim_1"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 23.0)
            np.testing.assert_equal(series[1], 24.0)

            series = df["dim_2"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_2"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 5.0)
            np.testing.assert_equal(series[1], 6.0)

            series = df["dim_2"][1]
            np.testing.assert_equal(len(series), 0)

            series = df["dim_2"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 25.0)
            np.testing.assert_equal(series[1], 26.0)

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and data that
    # contains datetimes as timestamps and has some empty dimensions is
    # classed as valid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel "
                "false\n@data\n"
            )
            file_contents += (
                "(01/01/2019 00:00:00, 1),  (01/02/2019 "
                "00:00:00, 2)  :                               "
                "                      : (01/05/2019 00:00:00, "
                "5), (01/06/2019 00:00:00, 6)\n"
            )
            file_contents += (
                "(01/01/2020 00:00:00, 11), (01/02/2020 "
                "00:00:00, 12) : (01/03/2020 00:00:00, 13), "
                "(01/04/2020 00:00:00, 14) :  \n"
            )
            file_contents += (
                "(01/01/2021 00:00:00, 21), (01/02/2021 "
                "00:00:00, 22) : (01/03/2021 00:00:00, 23), "
                "(01/04/2021 00:00:00, 24) :  \n"
            )

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file

            df = load_from_tsfile_to_dataframe(path)

            # Test the DataFrame returned accurately reflects the data in
            # the file

            np.testing.assert_equal(len(df), 3)
            np.testing.assert_equal(len(df.columns), 3)

            series = df["dim_0"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_0"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series["01/01/2019"], 1.0)
            np.testing.assert_equal(series["01/02/2019"], 2.0)

            series = df["dim_0"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series["01/01/2020"], 11.0)
            np.testing.assert_equal(series["01/02/2020"], 12.0)

            series = df["dim_0"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series["01/01/2021"], 21.0)
            np.testing.assert_equal(series["01/02/2021"], 22.0)

            series = df["dim_1"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_1"][0]
            np.testing.assert_equal(len(series), 0)

            series = df["dim_1"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series["01/03/2020"], 13.0)
            np.testing.assert_equal(series["01/04/2020"], 14.0)

            series = df["dim_1"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series["01/03/2021"], 23.0)
            np.testing.assert_equal(series["01/04/2021"], 24.0)

            series = df["dim_2"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_2"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series["01/05/2019"], 5.0)
            np.testing.assert_equal(series["01/06/2019"], 6.0)

            series = df["dim_2"][1]
            np.testing.assert_equal(len(series), 0)

            series = df["dim_2"][2]
            np.testing.assert_equal(len(series), 0)

    finally:
        os.remove(path)

    # Test that a file that mixes timestamp conventions is invalid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel "
                "false\n@data\n"
            )
            file_contents += (
                "(01/01/2019 00:00:00, 1),  (01/02/2019 "
                "00:00:00, 2)  :                               "
                "                      : (01/05/2019 00:00:00, "
                "5), (01/06/2019 00:00:00, 6)\n"
            )
            file_contents += (
                "(00, 11), (1, 12) : (01/03/2020 00:00:00, 13), "
                "(01/04/2020 00:00:00, 14) :  \n"
            )
            file_contents += (
                "(01/01/2021 00:00:00, 21), (01/02/2021 "
                "00:00:00, 22) : (01/03/2021 00:00:00, 23), "
                "(01/04/2021 00:00:00, 24) :  \n"
            )

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file and assert that it is invalid

            np.testing.assert_raises(
                TsFileParseException, load_from_tsfile_to_dataframe, path
            )

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and data but missing
    # classes is classed as invalid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel true 0 1 "
                "2\n@data\n"
            )
            file_contents += "(0, 1), (1, 2):(0, 3), (1, 4):(0, 5), (1, 6)\n"
            file_contents += "(0, 11), (1, 12):(0, 13), (1,14):(0, 15), (1, 16)     \n"

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file and assert that it is invalid

            np.testing.assert_raises(
                TsFileParseException, load_from_tsfile_to_dataframe, path
            )

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and data but invalid
    # classes is classed as invalid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel true 0 1 "
                "2\n@data\n"
            )
            file_contents += "(0, 1), (1, 2):(0, 3), (1, 4):(0, 5), (1, 6) : 0 \n"
            file_contents += (
                "(0, 11), (1, 12):(0, 13), (1,14):(0, 15), (1, 16)   : 3  \n"
            )

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file and assert that it is invalid

            np.testing.assert_raises(
                TsFileParseException, load_from_tsfile_to_dataframe, path
            )

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and data with classes
    # is classed as valid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "true\n@univariate true\n@classLabel true 0 1 "
                "2\n@data\n"
            )
            file_contents += "(0, 1), (1, 2):(0, 3), (1, 4):(0, 5), (1, 6): 0\n"
            file_contents += (
                "(0, 11), (1, 12):(0, 13), (1,14):(0, 15), (1, 16): 2     \n"
            )

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file

            df, y = load_from_tsfile_to_dataframe(path)

            # Test the DataFrame of X values returned accurately reflects
            # the data in the file

            np.testing.assert_equal(len(df), 2)
            np.testing.assert_equal(len(df.columns), 3)

            series = df["dim_0"]
            np.testing.assert_equal(len(series), 2)

            series = df["dim_0"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 1.0)
            np.testing.assert_equal(series[1], 2.0)

            series = df["dim_0"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 11.0)
            np.testing.assert_equal(series[1], 12.0)

            series = df["dim_1"]
            np.testing.assert_equal(len(series), 2)

            series = df["dim_1"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 3.0)
            np.testing.assert_equal(series[1], 4.0)

            series = df["dim_1"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 13.0)
            np.testing.assert_equal(series[1], 14.0)

            series = df["dim_2"]
            np.testing.assert_equal(len(series), 2)

            series = df["dim_2"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 5.0)
            np.testing.assert_equal(series[1], 6.0)

            series = df["dim_2"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 15.0)
            np.testing.assert_equal(series[1], 16.0)

            # Test that the class values are as expected

            np.testing.assert_equal(len(y), 2)
            np.testing.assert_equal(y[0], "0")
            np.testing.assert_equal(y[1], "2")

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and data, with no
    # timestamps, is classed as valid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "false\n@univariate true\n@classLabel "
                "false\n@data\n"
            )
            file_contents += "1,2:3,4:5,6\n"
            file_contents += "11,12:13,14:15,16\n"
            file_contents += "21,22:23,24:25,26\n"

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file

            df = load_from_tsfile_to_dataframe(path)

            # Test the DataFrame returned accurately reflects the data in
            # the file

            np.testing.assert_equal(len(df), 3)
            np.testing.assert_equal(len(df.columns), 3)

            series = df["dim_0"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_0"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 1.0)
            np.testing.assert_equal(series[1], 2.0)

            series = df["dim_0"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 11.0)
            np.testing.assert_equal(series[1], 12.0)

            series = df["dim_0"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 21.0)
            np.testing.assert_equal(series[1], 22.0)

            series = df["dim_1"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_1"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 3.0)
            np.testing.assert_equal(series[1], 4.0)

            series = df["dim_1"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 13.0)
            np.testing.assert_equal(series[1], 14.0)

            series = df["dim_1"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 23.0)
            np.testing.assert_equal(series[1], 24.0)

            series = df["dim_2"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_2"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 5.0)
            np.testing.assert_equal(series[1], 6.0)

            series = df["dim_2"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 15.0)
            np.testing.assert_equal(series[1], 16.0)

            series = df["dim_2"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 25.0)
            np.testing.assert_equal(series[1], 26.0)

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and data, with no
    # timestamps and some empty dimensions, is classed as valid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:

            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "false\n@univariate true\n@classLabel "
                "false\n@data\n"
            )
            file_contents += "1,2::5,6\n"
            file_contents += "11,12:13,14:15,16\n"
            file_contents += "21,22:23,24:\n"

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file

            df = load_from_tsfile_to_dataframe(path)

            # Test the DataFrame returned accurately reflects the data in
            # the file

            np.testing.assert_equal(len(df), 3)
            np.testing.assert_equal(len(df.columns), 3)

            series = df["dim_0"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_0"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 1.0)
            np.testing.assert_equal(series[1], 2.0)

            series = df["dim_0"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 11.0)
            np.testing.assert_equal(series[1], 12.0)

            series = df["dim_0"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 21.0)
            np.testing.assert_equal(series[1], 22.0)

            series = df["dim_1"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_1"][0]
            np.testing.assert_equal(len(series), 0)

            series = df["dim_1"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 13.0)
            np.testing.assert_equal(series[1], 14.0)

            series = df["dim_1"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 23.0)
            np.testing.assert_equal(series[1], 24.0)

            series = df["dim_2"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_2"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 5.0)
            np.testing.assert_equal(series[1], 6.0)

            series = df["dim_2"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 15.0)
            np.testing.assert_equal(series[1], 16.0)

            series = df["dim_2"][2]
            np.testing.assert_equal(len(series), 0)

    finally:
        os.remove(path)

    # Test that a file with a complete set of metadata and data, with no
    # timestamps and some empty dimensions and classes, is classed as valid

    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp_file:
            # Write the contents of the file

            file_contents = (
                "@problemName Test Problem\n@timeStamps "
                "false\n@univariate true\n@classLabel true cat "
                "bear dog\n@data\n"
            )
            file_contents += "1,2::5,6:cat  \n"
            file_contents += "11,12:13,14:15,16:  dog\n"
            file_contents += "21,22:23,24::   bear   \n"

            tmp_file.write(file_contents)
            tmp_file.flush()

            # Parse the file

            df, y = load_from_tsfile_to_dataframe(path)

            # Test the DataFrame of X values returned accurately reflects
            # the data in the file

            np.testing.assert_equal(len(df), 3)
            np.testing.assert_equal(len(df.columns), 3)

            series = df["dim_0"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_0"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 1.0)
            np.testing.assert_equal(series[1], 2.0)

            series = df["dim_0"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 11.0)
            np.testing.assert_equal(series[1], 12.0)

            series = df["dim_0"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 21.0)
            np.testing.assert_equal(series[1], 22.0)

            series = df["dim_1"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_1"][0]
            np.testing.assert_equal(len(series), 0)

            series = df["dim_1"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 13.0)
            np.testing.assert_equal(series[1], 14.0)

            series = df["dim_1"][2]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 23.0)
            np.testing.assert_equal(series[1], 24.0)

            series = df["dim_2"]
            np.testing.assert_equal(len(series), 3)

            series = df["dim_2"][0]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 5.0)
            np.testing.assert_equal(series[1], 6.0)

            series = df["dim_2"][1]
            np.testing.assert_equal(len(series), 2)
            np.testing.assert_equal(series[0], 15.0)
            np.testing.assert_equal(series[1], 16.0)

            series = df["dim_2"][2]
            np.testing.assert_equal(len(series), 0)

            # Test that the class values are as expected

            np.testing.assert_equal(len(y), 3)
            np.testing.assert_equal(y[0], "cat")
            np.testing.assert_equal(y[1], "dog")
            np.testing.assert_equal(y[2], "bear")

    finally:
        os.remove(path)


def test_load_from_long_to_dataframe(tmpdir):
    # create and save a example long-format file to csv
    test_dataframe = generate_example_long_table()
    dataframe_path = tmpdir.join("data.csv")
    test_dataframe.to_csv(dataframe_path, index=False)
    # load and convert the csv to sktime-formatted data
    nested_dataframe = load_from_long_to_dataframe(dataframe_path)
    assert isinstance(nested_dataframe, pd.DataFrame)


def test_load_from_long_incorrect_format(tmpdir):
    with pytest.raises(LongFormatDataParseException):
        dataframe = generate_example_long_table()
        dataframe.drop(dataframe.columns[[3]], axis=1, inplace=True)
        dataframe_path = tmpdir.join("data.csv")
        dataframe.to_csv(dataframe_path, index=False)
        load_from_long_to_dataframe(dataframe_path)



# -*- coding: utf-8 -*-
__author__ = "Markus Löning"
__all__ = ["all_estimators"]

import inspect
import pkgutil
from importlib import import_module
from operator import itemgetter
from pathlib import Path


def _get_name(obj):
    return obj.__class__.__name__


def all_estimators(estimator_types=None, return_names=True, exclude_estimators=None):
    """Get a list of all estimators from sktime.

    This function crawls the module and gets all classes that inherit
    from sktime's and sklearn's base classes.

    Not included are: the base classes themselves, classes defined in test
    modules.

    Parameters
    ----------
    estimator_types : string, list of string, optional (default=None)
        Which kind of estimators should be returned.
        - If None, no filter is applied and all estimators are returned.
        - Possible values are 'classifier', 'regressor', 'transformer' and
        'forecaster' to get estimators only of these specific types, or a list of
        these to get the estimators that fit at least one of the types.
    return_names : bool, optional (default=True)
        If True, return estimators as list of (name, estimator) tuples.
        If False, return list of estimators.
    exclude_estimators : str, list of str, optional (default=None)
        Names of estimators to exclude.

    Returns
    -------
    estimators : list of tuples
        List of (name, class), where ``name`` is the class name as string
        and ``class`` is the actual class.

    References
    ----------
    Modified version from scikit-learn's `all_estimators()`.
    """

    # lazy import to avoid circular imports
    import warnings
    from sktime.tests._config import (
        VALID_ESTIMATOR_TYPES,
        VALID_ESTIMATOR_BASE_TYPE_LOOKUP,
    )

    MODULES_TO_IGNORE = ("tests", "setup", "contrib", "benchmarking")
    all_estimators = []
    ROOT = str(Path(__file__).parent.parent)  # sktime package root directory

    def _is_abstract(klass):
        if not (hasattr(klass, "__abstractmethods__")):
            return False
        if not len(klass.__abstractmethods__):
            return False
        return True

    def _is_private_module(module):
        return "._" in module

    def _is_ignored_module(module):
        module_parts = module.split(".")
        return any(part in MODULES_TO_IGNORE for part in module_parts)

    def _is_base_class(name):
        return name.startswith("_") or name.startswith("Base")

    def _is_estimator(name, klass):
        # Check if klass is subclass of base estimators, not an base class itself and
        # not an abstract class
        return (
            issubclass(klass, VALID_ESTIMATOR_TYPES)
            and klass not in VALID_ESTIMATOR_TYPES
            and not _is_abstract(klass)
            and not _is_base_class(name)
        )

    # Ignore deprecation warnings triggered at import time and from walking
    # packages
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=FutureWarning)
        for _, module_name, _ in pkgutil.walk_packages(path=[ROOT], prefix="sktime."):

            # Filter modules
            if _is_ignored_module(module_name) or _is_private_module(module_name):
                continue

            module = import_module(module_name)
            classes = inspect.getmembers(module, inspect.isclass)

            # Filter classes
            estimators = [
                (name, klass) for name, klass in classes if _is_estimator(name, klass)
            ]
            all_estimators.extend(estimators)

    # Drop duplicates
    all_estimators = set(all_estimators)

    # Filter based on given estimator types
    def _is_in_estimator_types(estimator, estimator_types):
        return any(
            [
                issubclass(estimator, estimator_type)
                for estimator_type in estimator_types
            ]
        )

    def _check_estimator_types(estimator_types):
        if not isinstance(estimator_types, list):
            estimator_types = [estimator_types]  # make iterable

        def _get_err_msg(estimator_type):
            return (
                f"Parameter `estimator_type` must be None, a string or a list of "
                f"strings. Valid string values are: "
                f"{tuple(VALID_ESTIMATOR_BASE_TYPE_LOOKUP.keys())}, but found: "
                f"{repr(estimator_type)}"
            )

        for i, estimator_type in enumerate(estimator_types):
            if not isinstance(estimator_type, (type, str)):
                raise ValueError(
                    "Please specify `estimator_types` as a list of str or " "types."
                )
            if isinstance(estimator_type, str):
                if estimator_type not in VALID_ESTIMATOR_BASE_TYPE_LOOKUP.keys():
                    raise ValueError(_get_err_msg(estimator_type))
                estimator_type = VALID_ESTIMATOR_BASE_TYPE_LOOKUP[estimator_type]
                estimator_types[i] = estimator_type
            elif isinstance(estimator_type, type):
                pass
            else:
                raise ValueError(_get_err_msg(estimator_type))
        return estimator_types

    if estimator_types is not None:
        estimator_types = _check_estimator_types(estimator_types)
        all_estimators = [
            (name, estimator)
            for name, estimator in all_estimators
            if _is_in_estimator_types(estimator, estimator_types)
        ]

    # Filter based on given exclude list
    if exclude_estimators is not None:
        if not isinstance(exclude_estimators, list):
            exclude_estimators = [exclude_estimators]  # make iterable
        if not all([isinstance(estimator, str) for estimator in exclude_estimators]):
            raise ValueError(
                "Please specify `exclude_estimators` as a list of strings."
            )
        all_estimators = [
            (name, estimator)
            for name, estimator in all_estimators
            if name not in exclude_estimators
        ]

    # Drop duplicates, sort for reproducibility
    # itemgetter is used to ensure the sort does not extend to the 2nd item of
    # the tuple
    all_estimators = sorted(all_estimators, key=itemgetter(0))

    # Return with string names or only estimator classes
    if return_names:
        return all_estimators
    else:
        return [estimator for (name, estimator) in all_estimators]



# -*- coding: utf-8 -*-
# import numpy as np
# from numpy import testing
#
# from sktime.contrib.hybrid._catch22_forest_classifier import Catch22ForestClassifier
# from sktime.datasets import load_gunpoint, load_basic_motions
#
#
# def test_catch22_forest_classifier_on_gunpoint():
#     # load gunpoint data
#     X_train, y_train = load_gunpoint(split="train", return_X_y=True)
#     X_test, y_test = load_gunpoint(split="test", return_X_y=True)
#     indices = np.random.RandomState(0).permutation(10)
#
#     # train c22f
#     c22f = Catch22ForestClassifier(random_state=0)
#     c22f.fit(X_train.iloc[indices], y_train[indices])
#
#     # assert probabilities are the same
#     probas = c22f.predict_proba(X_test.iloc[indices])
#     testing.assert_array_equal(probas, catch22_forest_classifier_gunpoint_probas)
#
#
# def test_catch22_forest_classifier_on_basic_motions():
#     # load basic motions data
#     X_train, y_train = load_basic_motions(split="train", return_X_y=True)
#     X_test, y_test = load_basic_motions(split="test", return_X_y=True)
#     indices = np.random.RandomState(0).permutation(20)
#
#     # train c22f
#     c22f = Catch22ForestClassifier(random_state=0)
#     c22f.fit(X_train.iloc[indices], y_train[indices])
#
#     # assert probabilities are the same
#     probas = c22f.predict_proba(X_test.iloc[indices])
#     testing.assert_array_equal(probas, catch22_forest_classifier_basic_motions_probas)
#
#
# catch22_forest_classifier_gunpoint_probas = np.array(
#     [
#         [
#             0.07,
#             0.93,
#         ],
#         [
#             0.35,
#             0.65,
#         ],
#         [
#             0.35,
#             0.65,
#         ],
#         [
#             0.13,
#             0.87,
#         ],
#         [
#             0.11,
#             0.89,
#         ],
#         [
#             0.79,
#             0.21,
#         ],
#         [
#             0.05,
#             0.95,
#         ],
#         [
#             0.59,
#             0.41,
#         ],
#         [
#             0.56,
#             0.44,
#         ],
#         [
#             0.06,
#             0.94,
#         ],
#     ]
# )
# catch22_forest_classifier_basic_motions_probas = np.array(
#     [
#         [
#             0.99,
#             0.01,
#         ],
#         [
#             0.16,
#             0.84,
#         ],
#         [
#             0.35,
#             0.65,
#         ],
#         [
#             0.0,
#             1.0,
#         ],
#         [
#             0.99,
#             0.01,
#         ],
#         [
#             0.97,
#             0.03,
#         ],
#         [
#             0.15,
#             0.85,
#         ],
#         [
#             0.94,
#             0.06,
#         ],
#         [
#             0.15,
#             0.85,
#         ],
#         [
#             0.01,
#             0.99,
#         ],
#         [
#             0.04,
#             0.96,
#         ],
#         [
#             0.86,
#             0.14,
#         ],
#         [
#             0.01,
#             0.99,
#         ],
#         [
#             0.03,
#             0.97,
#         ],
#         [
#             0.43,
#             0.57,
#         ],
#         [
#             0.96,
#             0.04,
#         ],
#         [
#             0.22,
#             0.78,
#         ],
#         [
#             0.36,
#             0.64,
#         ],
#         [
#             0.35,
#             0.65,
#         ],
#         [
#             0.94,
#             0.06,
#         ],
#     ]
# )
#
#
# # def print_array(array):
# #     print('[')
# #     for sub_array in array:
# #         print('[')
# #         for value in sub_array:
# #             print(value.astype(str), end='')
# #             print(', ')
# #         print('],')
# #     print(']')
# #
# # if __name__ == "__main__":
# #     X_train, y_train = load_gunpoint(split="train", return_X_y=True)
# #     X_test, y_test = load_gunpoint(split="test", return_X_y=True)
# #     indices = np.random.RandomState(0).permutation(10)
# #
# #     c22f_u = Catch22ForestClassifier(random_state=0)
# #
# #     c22f_u.fit(X_train.iloc[indices], y_train[indices])
# #     probas = c22f_u.predict_proba(X_test.iloc[indices])
# #     print_array(probas)
# #
# #     X_train, y_train = load_basic_motions(split="train", return_X_y=True)
# #     X_test, y_test = load_basic_motions(split="test", return_X_y=True)
# #     indices = np.random.RandomState(0).permutation(20)
# #
# #     c22f_m = Catch22ForestClassifier(random_state=0)
# #
# #     c22f_m.fit(X_train.iloc[indices], y_train[indices])
# #     probas = c22f_m.predict_proba(X_test.iloc[indices])
# #     print_array(probas)



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Markus Löning"]
__all__ = ["BaseRegressor"]

from sktime.base import BaseEstimator


class BaseRegressor(BaseEstimator):
    """
    Base class for regressors, for identification.
    """

    def fit(self, X, y):
        raise NotImplementedError("abstract method")

    def predict(self, X):
        raise NotImplementedError("abstract method")

    def score(self, X, y):
        from sklearn.metrics import r2_score

        return r2_score(y, self.predict(X))



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__all__ = ["plot_series"]
__author__ = ["Markus Löning"]

import numpy as np

from sktime.utils.validation._dependencies import _check_soft_dependencies
from sktime.utils.validation.forecasting import check_y


def plot_series(*series, labels=None):
    """Plot one or more time series

    Parameters
    ----------
    series : pd.Series
        One or more time series
    labels : list, optional (default=None)
        Names of series, will be displayed in figure legend

    Returns
    -------
    fig : plt.Figure
    ax : plt.Axis
    """
    _check_soft_dependencies("matplotlib", "seaborn")
    import matplotlib.pyplot as plt
    from matplotlib.ticker import FuncFormatter, MaxNLocator
    from matplotlib.cbook import flatten
    import seaborn as sns

    n_series = len(series)
    if labels is not None:
        if n_series != len(labels):
            raise ValueError(
                "There must be one label for each time series, "
                "but found inconsistent numbers of series and "
                "labels."
            )
        legend = True
    else:
        labels = ["" for _ in range(n_series)]
        legend = False

    for y in series:
        check_y(y)

    # create combined index
    index = series[0].index
    for y in series[1:]:
        # check types, note that isinstance() does not work here because index
        # types inherit from each other, hence we check for type equality
        if not type(index) is type(y.index):  # noqa
            raise TypeError("Found series with different index types.")
        index = index.union(y.index)

    # generate integer x-values
    xs = [np.argwhere(index.isin(y.index)).ravel() for y in series]

    # create figure
    fig, ax = plt.subplots(1, figsize=plt.figaspect(0.25))
    colors = sns.color_palette("colorblind", n_colors=n_series)

    # plot series
    for x, y, color, label in zip(xs, series, colors, labels):

        # scatter if little data is available or index is not complete
        if len(x) <= 3 or not np.array_equal(np.arange(x[0], x[-1] + 1), x):
            plot_func = sns.scatterplot
        else:
            plot_func = sns.lineplot

        plot_func(x=x, y=y, ax=ax, marker="o", label=label, color=color)

    # combine data points for all series
    xs_flat = list(flatten(xs))

    # set x label of data point to the matching index
    def format_fn(tick_val, tick_pos):
        if int(tick_val) in xs_flat:
            return index[int(tick_val)]
        else:
            return ""

    # dynamically set x label ticks and spacing from index labels
    ax.xaxis.set_major_formatter(FuncFormatter(format_fn))
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))

    if legend:
        ax.legend()

    return fig, ax



# -*- coding: utf-8 -*-
__author__ = ["Jason Pong"]
__all__ = ["test_write_dataframe_to_ts_success", "test_write_dataframe_to_ts_fail"]

import os

import numpy as np
import pytest
from pandas._testing import assert_frame_equal

import sktime

from sktime.utils.data_io import load_from_tsfile_to_dataframe
from sktime.utils.data_io import write_dataframe_to_tsfile


def test_write_dataframe_to_ts_success(tmp_path):
    # load an example dataset
    path = os.path.join(
        os.path.dirname(sktime.__file__),
        "datasets/data/ItalyPowerDemand/ItalyPowerDemand_TEST.ts",
    )
    test_x, test_y = load_from_tsfile_to_dataframe(path)
    # output the dataframe in a ts file
    write_dataframe_to_tsfile(
        data=test_x,
        path=tmp_path,
        problem_name="ItalyPowerDemand",
        timestamp=False,
        univariate=True,
        class_label=[1, 2],
        class_value_list=test_y,
        comment="""
          The data was derived from twelve monthly electrical power demand
          time series from Italy and first used in the paper "Intelligent
          Icons: Integrating Lite-Weight Data Mining and Visualization into
          GUI Operating Systems". The classification task is to distinguish
          days from Oct to March (inclusive) from April to September.
        """,
    )
    # load data back from the ts file
    result = f"{tmp_path}/ItalyPowerDemand/ItalyPowerDemand_transform.ts"
    res_x, res_y = load_from_tsfile_to_dataframe(result)
    # check if the dataframes are the same
    assert_frame_equal(res_x, test_x)


def test_write_dataframe_to_ts_fail(tmp_path):
    with pytest.raises(ValueError, match="Data provided must be a DataFrame"):
        write_dataframe_to_tsfile(
            data=np.random.rand(3, 2),
            path=str(tmp_path),
            problem_name="GunPoint",
            timestamp=False,
            univariate=True,
        )



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Markus Löning"]
__all__ = [
    "Deseasonalizer",
    "Deseasonalizer",
    "ConditionalDeseasonalizer",
    "ConditionalDeseasonalizer",
]

import numpy as np
from statsmodels.tsa.seasonal import seasonal_decompose

from sktime.transformations.base import _SeriesToSeriesTransformer
from sktime.utils.datetime import _get_duration
from sktime.utils.datetime import _get_unit
from sktime.utils.seasonality import autocorrelation_seasonality_test
from sktime.utils.validation.forecasting import check_sp
from sktime.utils.validation.series import check_series


class Deseasonalizer(_SeriesToSeriesTransformer):
    """A transformer that removes a seasonal and trend components from time
    series

    Parameters
    ----------
    sp : int, optional (default=1)
        Seasonal periodicity
    model : str {"additive", "multiplicative"}, optional (default="additive")
        Model to use for estimating seasonal component
    """

    _tags = {"transform-returns-same-time-index": True, "univariate-only": True}

    def __init__(self, sp=1, model="additive"):
        self.sp = check_sp(sp)
        allowed_models = ("additive", "multiplicative")
        if model not in allowed_models:
            raise ValueError(
                f"`model` must be one of {allowed_models}, " f"but found: {model}"
            )
        self.model = model
        self._y_index = None
        self.seasonal_ = None
        super(Deseasonalizer, self).__init__()

    def _set_y_index(self, y):
        self._y_index = y.index

    def _align_seasonal(self, y):
        """Align seasonal components with y's time index"""
        shift = (
            -_get_duration(
                y.index[0],
                self._y_index[0],
                coerce_to_int=True,
                unit=_get_unit(self._y_index),
            )
            % self.sp
        )
        return np.resize(np.roll(self.seasonal_, shift=shift), y.shape[0])

    def fit(self, Z, X=None):
        """Fit to data.

        Parameters
        ----------
        Z : pd.Series
        X : pd.DataFrame

        Returns
        -------
        self : an instance of self
        """
        z = check_series(Z, enforce_univariate=True)
        self._set_y_index(z)
        sp = check_sp(self.sp)

        # apply seasonal decomposition
        self.seasonal_ = seasonal_decompose(
            z,
            model=self.model,
            period=sp,
            filt=None,
            two_sided=True,
            extrapolate_trend=0,
        ).seasonal.iloc[:sp]

        self._is_fitted = True
        return self

    def _transform(self, y, seasonal):
        if self.model == "additive":
            return y - seasonal
        else:
            return y / seasonal

    def _inverse_transform(self, y, seasonal):
        if self.model == "additive":
            return y + seasonal
        else:
            return y * seasonal

    def transform(self, Z, X=None):
        """Transform data.
        Returns a transformed version of y.

        Parameters
        ----------
        y : pd.Series
        X : pd.DataFrame

        Returns
        -------
        yt : pd.Series
            Transformed time series.
        """
        self.check_is_fitted()
        z = check_series(Z, enforce_univariate=True)
        seasonal = self._align_seasonal(z)
        return self._transform(z, seasonal)

    def inverse_transform(self, Z, X=None):
        """Inverse transform data.
        Returns a transformed version of y.

        Parameters
        ----------
        y : pd.Series
        X : pd.DataFrame

        Returns
        -------
        yt : pd.Series
            Transformed time series.
        """
        self.check_is_fitted()
        z = check_series(Z, enforce_univariate=True)
        seasonal = self._align_seasonal(z)
        return self._inverse_transform(z, seasonal)

    def update(self, Z, X=None, update_params=False):
        """Update fitted parameters

        Parameters
        ----------
        y : pd.Series
        X : pd.DataFrame
        update_params : bool, optional (default=False)

        Returns
        -------
        self : an instance of self
        """
        self.check_is_fitted()
        z = check_series(Z, enforce_univariate=True)
        self._set_y_index(z)
        return self


class ConditionalDeseasonalizer(Deseasonalizer):
    """A transformer that removes a seasonal and trend components from time
    series, conditional on seasonality test.

    Parameters
    ----------
    seasonality_test : callable, optional (default=None)
        Callable that tests for seasonality and returns True when data is
        seasonal and False otherwise. If None,
        90% autocorrelation seasonality test is used.
    sp : int, optional (default=1)
        Seasonal periodicity
    model : str {"additive", "multiplicative"}, optional (default="additive")
        Model to use for estimating seasonal component
    """

    def __init__(self, seasonality_test=None, sp=1, model="additive"):
        self.seasonality_test = seasonality_test
        self.is_seasonal_ = None
        super(ConditionalDeseasonalizer, self).__init__(sp=sp, model=model)

    def _check_condition(self, y):
        """Check if y meets condition"""

        if not callable(self.seasonality_test_):
            raise ValueError(
                f"`func` must be a function/callable, but found: "
                f"{type(self.seasonality_test_)}"
            )

        is_seasonal = self.seasonality_test_(y, sp=self.sp)
        if not isinstance(is_seasonal, (bool, np.bool_)):
            raise ValueError(
                f"Return type of `func` must be boolean, "
                f"but found: {type(is_seasonal)}"
            )
        return is_seasonal

    def fit(self, Z, X=None):
        """Fit to data.

        Parameters
        ----------
        y_train : pd.Series

        Returns
        -------
        self : an instance of self
        """

        z = check_series(Z, enforce_univariate=True)
        self._set_y_index(z)
        sp = check_sp(self.sp)

        # set default condition
        if self.seasonality_test is None:
            self.seasonality_test_ = autocorrelation_seasonality_test
        else:
            self.seasonality_test_ = self.seasonality_test

        # check if data meets condition
        self.is_seasonal_ = self._check_condition(z)

        if self.is_seasonal_:
            # if condition is met, apply de-seasonalisation
            self.seasonal_ = seasonal_decompose(
                z,
                model=self.model,
                period=sp,
                filt=None,
                two_sided=True,
                extrapolate_trend=0,
            ).seasonal.iloc[:sp]
        else:
            # otherwise, set idempotent seasonal components
            self.seasonal_ = (
                np.zeros(self.sp) if self.model == "additive" else np.ones(self.sp)
            )

        self._is_fitted = True
        return self



# -*- coding: utf-8 -*-
import numpy as np
from scipy.optimize import bisect
from scipy.optimize import nnls


class _PredictionWeightedEnsembler:
    """Wrapper class to handle ensemble algorithms that use multiple forecasters
    for prediction. We implement default methods for setting uniform weights,
    updating and prediction.

    Parameters
    ----------
    n_estimators : float
        number of estimators
    loss_func : function
        loss function which follows sklearn.metrics API, for updating weights
    """

    def __init__(self, n_estimators=10, loss_func=None):
        self.n_estimators = n_estimators
        self.weights = np.ones(n_estimators) / n_estimators
        self.loss_func = loss_func
        super(_PredictionWeightedEnsembler, self).__init__()

    def predict(self, y_pred):
        """Performs prediction by taking a weighted average of the estimator
            predictions w.r.t the weights vector

        Parameters
        ----------
        y_pred : np.array(), shape=(time_axis,estimator_axis)
            array with predictions from the estimators

        Returns
        -------
        predictions : np.array(), shape=(time_axis)
            array with our predictions
        """
        prediction = np.dot(self.weights, y_pred)

        return prediction

    def _modify_weights(self, new_array):
        """Performs a pointwise multiplication of the current
        weights with a new array of weights.

        Parameters
        ----------
        new_array : np.array()
            input array for pointwise multiplication
        """
        self.weights = self.weights * new_array
        self.weights /= np.sum(self.weights)

    def update(self, y_pred, y_true):
        """Resets the weights over the estimators by passing previous observations
            to the weighting algorithm

        Parameters
        ----------
        y_pred : np.array(), shape=(time_axis,estimator_axis)
            array with predictions from the estimators
        y_true : np.array(), shape=(time_axis)
            array with actual values for predicted quantity
        """
        raise NotImplementedError()

    def _uniform_weights(self, n_estimators):
        """Resets weights for n estimator to uniform weights

        Parameters
        ----------
        n_estimators : int
            number of estimators
        """
        self.n = n_estimators
        self.weights = np.ones(n_estimators) / n_estimators


class HedgeExpertEnsemble(_PredictionWeightedEnsembler):
    """Wrapper class to set parameters for hedge-style ensemble algorithms with
    a forecasting horizon and normalizing constant.

    Parameters
    ----------
    n_estimators : float
        number of estimators
    T : int
        forecasting horizon (in terms of timesteps)
    a : float
        normalizing constant
    loss_func : function
        loss function which follows sklearn.metrics API, for updating weights
    """

    def __init__(self, n_estimators=10, T=10, a=1, loss_func=None):
        super().__init__(n_estimators=n_estimators, loss_func=loss_func)
        self.T = T
        self.a = a
        self._uniform_weights(n_estimators)
        self.loss_func = loss_func


class NormalHedgeEnsemble(HedgeExpertEnsemble):
    """Implementation of A Parameter-free Hedging Algorithm,
    Kamalika Chaudhuri, Yoav Freund, Daniel Hsu (2009) as a hedge-style
    algorithm.

    Parameters
    ----------
    n_estimators : float
        number of estimators
    T : int
        forecasting horizon (in terms of timesteps)
    a : float
        normalizing constant
    loss_func : function
        loss function which follows sklearn.metrics API, for updating weights
    """

    def __init__(self, n_estimators=10, a=1, loss_func=None):
        super().__init__(n_estimators=n_estimators, T=None, a=a, loss_func=loss_func)
        self.R = np.zeros(n_estimators)

    def update(self, y_pred, y_true, low_c=0.01):
        """Resets the weights over the estimators by passing previous observations
            and updating based on Normal Hedge.

        Parameters
        ----------
        y_pred : np.array(), shape=(time_axis,estimator_axis)
            array with predictions from the estimators
        y_true : np.array(), shape=(time_axis)
            array with actual values for predicted quantity
        """
        assert y_pred.shape[1] == len(y_true), "Time Dimension Matches"
        time_length = y_pred.shape[1]

        for i in range(time_length):
            loss_vector = np.array(
                [
                    self.loss_func([prediction], [y_true[i]])
                    for prediction in y_pred[:, i]
                ]
            )

            average_loss = np.dot(self.weights, loss_vector)

            instant_regret = average_loss - loss_vector
            self.R += instant_regret
            self._update_weights(low_c=low_c)

    def _update_weights(self, low_c=0.01):
        """Updates the weights on each of the estimators by performing a potential
        function update with a root-finding search. low_c represents the lower
        bound on the window that the root finding is occuring over.

        Parameters
        ----------
        low_c : float
            lowest value that c can take
        """

        # Calculating Normalizing Constant
        R_plus = np.array(list(map(lambda x: 0 if 0 > x else x, self.R)))
        normalizing_R = np.max(R_plus)

        R_plus /= normalizing_R

        low_c = low_c
        high_c = (max(R_plus) ** 2) / 2

        def _pot(c):
            """Internal Potential Function

            Parameters
            ----------
            low_c : float
                lowest value that c can take

            Returns
            -------
            potential: float
            """
            return np.mean(np.exp((R_plus ** 2) / (2 * c))) - np.e

        c_t = bisect(_pot, low_c, high_c)

        def _prob(r, c_t):
            """Internal Probability Function

            Parameters
            ----------
            r : float
                regret
            c_t : float
                current value for c

            Returns
            -------
            prob : float
                probability
            """
            return (r / c_t) * np.exp((r ** 2) / (2 * c_t))

        self.weights = np.array([_prob(r, c_t) for r in R_plus])
        self.weights /= np.sum(self.weights)


class NNLSEnsemble(_PredictionWeightedEnsembler):
    """Ensemble class that performs a non-negative least squares to fit to the
    estimators. Keeps track of all observations seen so far and fits to it.

    Parameters
    ----------
    n_estimators: int
        number of estimators
    loss_func : function
        loss function which follows sklearn.metrics API, for updating weights
    """

    def __init__(self, n_estimators=10, loss_func=None):
        super().__init__(n_estimators=n_estimators, loss_func=loss_func)
        self.total_y_pred = np.empty((n_estimators, 0))
        self.total_y_true = np.empty(0)

    def update(self, y_pred, y_true):
        self.total_y_pred = np.concatenate((self.total_y_pred, y_pred), axis=1)
        self.total_y_true = np.concatenate((self.total_y_true, y_true))
        weights, loss = nnls(self.total_y_pred.T, self.total_y_true)
        self.weights = weights



# -*- coding: utf-8 -*-
""" Shapelet Transform Classifier
wrapper implementation of a shapelet transform classifier pipeline that
simply performs a (configurable) shapelet transform
then builds (by default) a random forest. This is a stripped down version
for basic usage

"""

__author__ = "Tony Bagnall"
__all__ = ["ShapeletTransformClassifier"]

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.pipeline import Pipeline
from sklearn.utils.multiclass import class_distribution
from sktime.classification.base import BaseClassifier
from sktime.transformations.panel.shapelets import ContractedShapeletTransform
from sktime.utils.validation.panel import check_X, check_X_y


class ShapeletTransformClassifier(BaseClassifier):
    """Shapelet Transform Classifier
        Basic implementation along the lines of [1,2]

    Parameters
    ____________
    TO DO

    Attributes
    ----------
    TO DO

    Notes
    _____

    ..[1] Jon Hills wt al., "Classification of time series by
    shapelet transformation",
        Data Mining and Knowledge Discovery, 28(4), 851--881, 2014
    https://link.springer.com/article/10.1007/s10618-013-0322-1
    ..[2] A. Bostrom and A. Bagnall, "Binary Shapelet Transform
    for Multiclass Time Series Classification",
    Transactions on Large-Scale Data and Knowledge Centered
      Systems, 32, 2017
    https://link.springer.com/chapter/10.1007/978-3-319-22729-0_20
    Java Version
    https://github.com/uea-machine-learning/tsml/blob/master/src/main/
    java/tsml/classifiers/shapelet_based/ShapeletTransformClassifier.java


    """

    def __init__(self, time_contract_in_mins=300, n_estimators=500, random_state=None):
        self.time_contract_in_mins = time_contract_in_mins
        self.n_estimators = n_estimators
        self.random_state = random_state

        #        self.shapelet_transform=ContractedShapeletTransform(
        #        time_limit_in_mins=self.time_contract_in_mins, verbose=shouty)
        #        self.classifier=RandomForestClassifier(
        #        n_estimators=self.n_estimators,criterion="entropy")
        #        self.st_X=None;
        super(ShapeletTransformClassifier, self).__init__()

    def fit(self, X, y):
        """Perform a shapelet transform then builds a random forest.
        Contract default for ST is 5 hours
        ----------
        X : array-like or sparse matrix of shape = [n_instances,
        series_length] or shape = [n_instances,n_columns]
            The training input samples.  If a Pandas data frame is passed it
            must have a single column (i.e. univariate
            classification. RISE has no bespoke method for multivariate
            classification as yet.
        y : array-like, shape =  [n_instances]    The class labels.

        Returns
        -------
        self : object
        """
        X, y = check_X_y(X, y, enforce_univariate=True)

        # if y is a pd.series then convert to array.
        if isinstance(y, pd.Series):
            y = y.to_numpy()

        # generate pipeline in fit so that random state can be propogated properly.
        self.classifier_ = Pipeline(
            [
                (
                    "st",
                    ContractedShapeletTransform(
                        time_contract_in_mins=self.time_contract_in_mins,
                        verbose=False,
                        random_state=self.random_state,
                    ),
                ),
                (
                    "rf",
                    RandomForestClassifier(
                        n_estimators=self.n_estimators, random_state=self.random_state
                    ),
                ),
            ]
        )

        self.n_classes_ = np.unique(y).shape[0]
        self.classes_ = class_distribution(np.asarray(y).reshape(-1, 1))[0][0]

        self.classifier_.fit(X, y)

        self._is_fitted = True
        return self

    def predict(self, X):
        """
        Find predictions for all cases in X. Built on top of predict_proba
        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samps, num_atts] or a
        data frame.
        If a Pandas data frame is passed,

        Returns
        -------
        output : array of shape = [n_samples]
        """
        X = check_X(X, enforce_univariate=True)
        self.check_is_fitted()

        return self.classifier_.predict(X)

    def predict_proba(self, X):
        """
        Find probability estimates for each class for all cases in X.
        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_instances, n_columns]
            The training input samples.  If a Pandas data frame is passed,

        Returns
        -------
        output : array of shape = [n_samples, num_classes] of probabilities
        """
        X = check_X(X, enforce_univariate=True)
        self.check_is_fitted()

        return self.classifier_.predict_proba(X)



# -*- coding: utf-8 -*-
__all__ = ["BaseClassifier"]
__author__ = ["Markus Löning"]

import numpy as np

from sktime.base import BaseEstimator
from sktime.utils.validation.panel import check_X


class BaseClassifier(BaseEstimator):
    """
    Base class for classifiers, for identification.
    """

    def fit(self, X, y):
        raise NotImplementedError("abstract method")

    def predict_proba(self, X):
        raise NotImplementedError("abstract method")

    def predict(self, X):
        """
        classify instances
        ----
        Parameters
        ----
        X : panda dataframe
            instances of the dataset
        ----
        Returns
        ----
        predictions : 1d numpy array
            array of predictions of each instance (class value)
        """
        X = check_X(X)
        self.check_is_fitted()
        distributions = self.predict_proba(X)
        predictions = []
        for instance_index in range(0, X.shape[0]):
            distribution = distributions[instance_index]
            prediction = np.argmax(distribution)
            predictions.append(prediction)
        predictions = self.label_encoder.inverse_transform(predictions)
        return predictions

    def score(self, X, y):
        from sklearn.metrics import accuracy_score

        return accuracy_score(y, self.predict(X), normalize=True)



# -*- coding: utf-8 -*-
import pandas as pd

from sktime.datasets import load_basic_motions
from sktime.transformations.panel.interpolate import TSInterpolator


def cut_X_ts(X):
    for row_i in range(X.shape[0]):
        for dim_i in range(X.shape[1]):
            ts = X.at[row_i, f"dim_{dim_i}"]
            X.at[row_i, f"dim_{dim_i}"] = pd.Series(ts.tolist()[: len(ts) - dim_i - 1])


def test_resizing():
    # 1) all lengths are equal
    # 2) cut lengths and check that they are really different
    # 3) use transformer for resizing to resize time series to equal length
    # 4) check that result length are equal to length that was set for
    #       transformer

    X, _ = load_basic_motions(split="train", return_X_y=True)

    # 1) Check that lengths of all time series (all via the axis=1 - for
    # all dims in first row) are equal.
    ts_lens_before = [len(X.iloc[0][i]) for i in range(len(X.iloc[0]))]
    # all lengths are equal to first length in array
    assert all([length == ts_lens_before[0] for length in ts_lens_before])

    # 2) cutting each time series in each cell of X to make lengths different
    cut_X_ts(X)  # operation is inplace
    # get lengths to ensure that they are really different
    ts_lens_after_cut = [len(X.iloc[0][i]) for i in range(len(X.iloc[0]))]
    assert not all(
        [length == ts_lens_after_cut[0] for length in ts_lens_after_cut]
    )  # are different

    # 3) make tranformer, set target length `target_len` and apply it
    target_len = 50
    Xt = TSInterpolator(target_len).fit_transform(X)

    # 4) check that result time series have lengths equal to `target_len
    #       that we set above
    ts_lens_after_resize = [len(Xt.iloc[0][i]) for i in range(len(Xt.iloc[0]))]
    assert all([length == target_len for length in ts_lens_after_resize])



# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Markus Löning"]
__all__ = ["ForecastingHorizon"]

from functools import lru_cache

import numpy as np
import pandas as pd

from sktime.utils.datetime import _coerce_duration_to_int
from sktime.utils.datetime import _get_unit
from sktime.utils.datetime import _shift
from sktime.utils.validation.series import VALID_INDEX_TYPES

RELATIVE_TYPES = (pd.Int64Index, pd.RangeIndex)
ABSOLUTE_TYPES = (pd.Int64Index, pd.RangeIndex, pd.DatetimeIndex, pd.PeriodIndex)
assert set(RELATIVE_TYPES).issubset(VALID_INDEX_TYPES)
assert set(ABSOLUTE_TYPES).issubset(VALID_INDEX_TYPES)

DELEGATED_METHODS = (
    "__sub__",
    "__add__",
    "__mul__",
    "__div__",
    "__divmod__",
    "__pow__",
    "__gt__",
    "__ge__",
    "__ne__",
    "__lt__",
    "__eq__",
    "__le__",
    "__radd__",
    "__rsub__",
    "__rmul__",
    "__rdiv__",
    "__rmod__",
    "__rdivmod__",
    "__rpow__",
    "__getitem__",
    "__len__",
    "max",
    "min",
)


def _delegator(method):
    """Helper function to automatically decorate ForecastingHorizon class with
    methods from pandas.Index and delegate method calls to wrapped pandas.Index
    object."""

    def delegated(obj, *args, **kwargs):
        return getattr(obj.to_pandas(), method)(*args, **kwargs)

    return delegated


def _check_values(values):
    """Validate forecasting horizon values and converts them to supported
    pandas.Index types if possible.

    Parameters
    ----------
    values : int, list, array, certain pd.Index types
        Forecasting horizon with steps ahead to predict.

    Raises
    ------
    TypeError : if values type is not supported

    Returns
    -------
    values : pd.Index
        Sorted and validated forecasting horizon values.
    """
    # if values are one of the supported pandas index types, we don't have
    # to do
    # anything as the forecasting horizon directly wraps the index, note that
    # isinstance() does not work here, because index types inherit from each
    # other,
    # hence we check for type equality here
    if type(values) in VALID_INDEX_TYPES:
        pass

    # convert single integer to pandas index, no further checks needed
    elif isinstance(values, (int, np.integer)):
        return pd.Int64Index([values], dtype=np.int)

    # convert np.array or list to pandas index
    elif isinstance(values, (list, np.ndarray)):
        values = pd.Int64Index(values, dtype=np.int)

    # otherwise, raise type error
    else:
        valid_types = (
            "int",
            "np.array",
            "list",
            *[f"pd.{index_type.__name__}" for index_type in VALID_INDEX_TYPES],
        )
        raise TypeError(
            f"`values` type not supported. `values` must be one of"
            f" {valid_types}, but found: {type(values)}"
        )

    # check values does not contain duplicates
    if len(values) != values.nunique():
        raise ValueError("`values` must not contain duplicates.")

    # return sorted values
    return values.sort_values()


class ForecastingHorizon:
    """Forecasting horizon

    Parameters
    ----------
    values : pd.Index, np.array, list or int
        Values of forecasting horizon
    is_relative : bool, optional (default=True)
        - If True, values are relative to end of training series.
        - If False, values are absolute.
    """

    def __new__(cls, values=None, is_relative=True):
        # We want the ForecastingHorizon class to be an extension of the
        # pandas index, but since subclassing pandas indices is not
        # straightforward, we wrap the index object instead. In order to
        # still support the basic methods of a pandas index, we dynamically
        # add some basic methods and delegate the method calls to the wrapped
        # index object.
        for method in DELEGATED_METHODS:
            setattr(cls, method, _delegator(method))
        return object.__new__(cls)

    def __init__(self, values=None, is_relative=True):
        if not isinstance(is_relative, bool):
            raise TypeError("`is_relative` must be a boolean")
        values = _check_values(values)

        # check types, note that isinstance() does not work here because index
        # types inherit from each other, hence we check for type equality
        error_msg = (
            f"`values` type is not compatible with `is_relative=" f"{is_relative}`."
        )
        if is_relative:
            if not type(values) in RELATIVE_TYPES:
                raise TypeError(error_msg)
        else:
            if not type(values) in ABSOLUTE_TYPES:
                raise TypeError(error_msg)

        self._values = values
        self._is_relative = is_relative

    def _new(self, values=None, is_relative=None):
        """Construct new ForecastingHorizon based on current object"""
        if values is None:
            values = self._values
        if is_relative is None:
            is_relative = self.is_relative
        return type(self)(values, is_relative)

    @property
    def is_relative(self):
        """Whether forecasting horizon is relative to the end of the
        training series.

        Returns
        -------
        is_relative : bool
        """
        return self._is_relative

    def to_pandas(self):
        """Returns underlying values as pd.Index

        Returns
        -------
        fh : pd.Index
        """
        return self._values

    def to_numpy(self, **kwargs):
        """Returns underlying values as np.array

        Parameters
        ----------
        **kwargs : dict of kwargs
            kwargs passed to `to_numpy()` of wrapped pandas index.

        Returns
        -------
        fh : np.ndarray
        """
        return self.to_pandas().to_numpy(**kwargs)

    # we cache the results from `to_relative()` and `to_absolute()` calls to speed up
    # computations, as these are the basic methods and often required internally when
    # calling different methods
    @lru_cache(typed=True)
    def to_relative(self, cutoff=None):
        """Return relative values
        Parameters
        ----------
        cutoff : pd.Period, pd.Timestamp, int, optional (default=None)
            Cutoff value is required to convert a relative forecasting
            horizon to an absolute one and vice versa.

        Returns
        -------
        fh : ForecastingHorizon
            Relative representation of forecasting horizon
        """
        if self.is_relative:
            return self._new()

        else:
            self._check_cutoff(cutoff)
            values = self.to_pandas() - cutoff

            if isinstance(self.to_pandas(), (pd.PeriodIndex, pd.DatetimeIndex)):
                values = _coerce_duration_to_int(values, unit=_get_unit(cutoff))

            return self._new(values, is_relative=True)

    @lru_cache(typed=True)
    def to_absolute(self, cutoff=None):
        """Return absolute values
        Parameters
        ----------
        cutoff : pd.Period, pd.Timestamp, int, optional (default=None)
            Cutoff value is required to convert a relative forecasting
            horizon to an absolute one and vice versa.

        Returns
        -------
        fh : ForecastingHorizon
            Absolute representation of forecasting horizon
        """
        if not self.is_relative:
            return self._new()

        else:
            self._check_cutoff(cutoff)
            index = self.to_pandas()
            values = _shift(cutoff, by=index)
            return self._new(values, is_relative=False)

    def to_absolute_int(self, start, cutoff=None):
        """Return absolute values as zero-based integer index
        Parameters
        ----------
        start : pd.Period, pd.Timestamp, int
            Start value
        cutoff : pd.Period, pd.Timestamp, int, optional (default=None)
            Cutoff value is required to convert a relative forecasting
            horizon to an absolute one and vice versa.
        Returns
        -------
        fh : ForecastingHorizon
            Absolute representation of forecasting horizon as zero-based
            integer index
        """
        self._check_cutoff(start)
        absolute = self.to_absolute(cutoff).to_pandas()
        values = absolute - start
        if isinstance(absolute, (pd.PeriodIndex, pd.DatetimeIndex)):
            values = _coerce_duration_to_int(values, unit=_get_unit(cutoff))
        return self._new(values, is_relative=False)

    def to_in_sample(self, cutoff=None):
        """Return in-sample values
        Parameters
        ----------
        cutoff : pd.Period, pd.Timestamp, int, optional (default=None)
            Cutoff value is required to convert a relative forecasting
            horizon to an absolute one and vice versa.

        Returns
        -------
        fh : ForecastingHorizon
            In-sample values of forecasting horizon
        """
        is_in_sample = self._is_in_sample(cutoff)
        in_sample = self.to_pandas()[is_in_sample]
        return self._new(in_sample)

    def to_out_of_sample(self, cutoff=None):
        """Return out-of-sample values
        Parameters
        ----------
        cutoff : pd.Period, pd.Timestamp, int, optional (default=None)
            Cutoff value is required to convert a relative forecasting
            horizon to an absolute one and vice versa.

        Returns
        -------
        fh : ForecastingHorizon
            Out-of-sample values of forecasting horizon
        """
        is_out_of_sample = self._is_out_of_sample(cutoff)
        out_of_sample = self.to_pandas()[is_out_of_sample]
        return self._new(out_of_sample)

    def _is_in_sample(self, cutoff=None):
        """Get index location of in-sample values"""
        return self.to_relative(cutoff).to_pandas() <= 0

    def is_all_in_sample(self, cutoff=None):
        """Whether or not the forecasting horizon is purely in-sample for given
        cutoff.

        Parameters
        ----------
        cutoff : pd.Period, pd.Timestamp, int, optional (default=None)
            Cutoff value is required to convert a relative forecasting
            horizon to an absolute one and vice versa.

        Returns
        -------
        ret : bool
            True if the forecasting horizon is purely in-sample for given cutoff.
        """
        return sum(self._is_in_sample(cutoff)) == len(self)

    def _is_out_of_sample(self, cutoff=None):
        """Get index location of out-of-sample values"""
        # return ~self._in_sample_idx(cutoff)
        return self.to_relative(cutoff).to_pandas() > 0

    def is_all_out_of_sample(self, cutoff=None):
        """Whether or not the forecasting horizon is purely out-of-sample for
        given cutoff.

        Parameters
        ----------
        cutoff : pd.Period, pd.Timestamp, int, optional (default=None)
            Cutoff value is required to convert a relative forecasting
            horizon to an absolute one and vice versa.

        Returns
        -------
        ret : bool
            True if the forecasting horizon is purely out-of-sample for given
            cutoff.
        """
        return sum(self._is_out_of_sample(cutoff)) == len(self)

    def to_indexer(self, cutoff=None, from_cutoff=True):
        """Return zero-based indexer values for easy indexing into arrays.
        Parameters
        ----------
        cutoff : pd.Period, pd.Timestamp, int, optional (default=None)
            Cutoff value is required to convert a relative forecasting
            horizon to an absolute one and vice versa.
        from_cutoff : bool, optional (default=True)
            - If True, zero-based relative to cutoff.
            - If False, zero-based relative to first value in forecasting
            horizon.

        Returns
        -------
        fh : pd.Index
            Indexer
        """
        if from_cutoff:
            return self.to_relative(cutoff).to_pandas() - 1
        else:
            relative = self.to_relative(cutoff)
            return relative - relative.to_pandas()[0]

    def __repr__(self):
        # generate repr based on wrapped index repr
        class_name = self.__class__.__name__
        pandas_repr = repr(self.to_pandas()).split("(")[-1].strip(")")
        return f"{class_name}({pandas_repr}, is_relative={self.is_relative})"

    def _check_cutoff(self, cutoff):
        """Helper function to check fh type compatibility against cutoff"""
        if cutoff is None:
            raise ValueError("`cutoff` must be provided.")

        index = self.to_pandas()

        if isinstance(index, pd.PeriodIndex):
            assert isinstance(cutoff, pd.Period)
        if isinstance(index, pd.DatetimeIndex):
            assert isinstance(cutoff, pd.Timestamp)



import numpy as np

__author__ = "Jason Lines"
"""
python versions of distance functions. These are fearful slow, and the cython ones should be used
these are in sktime.distances.elastic

        """


def dtw_distance(first, second, **kwargs):

    def dtw_single_channel(first, second, **kwargs):
        cutoff = np.inf
        try:
            window = kwargs["window"]
        except:
            window = 1.0
        # print("window: "+str(window))

        n = len(first)
        m = len(second)

        warp_matrix = np.full([n, m], np.inf)
        if n > m:
            window_size = n * window
        else:
            window_size = m * window
        window_size = int(window_size)

        dist = lambda x1, x2: ((x1 - x2) ** 2)

        pairwise_distances = np.asarray([[dist(x1, x2) for x2 in second] for x1 in first])

        # initialise edges of the warping matrix
        warp_matrix[0][0] = pairwise_distances[0][0]
        for i in range(1, window_size):
            warp_matrix[0][i] = pairwise_distances[0][i] + warp_matrix[0][i - 1]
            warp_matrix[i][0] = pairwise_distances[i][0] + warp_matrix[i - 1][0]

        # now visit all allowed cells, calculate the value as the distance in this cell + min(top, left, or top-left)
        # traverse each row,
        for row in range(1, n):
            cutoff_beaten = False
            # traverse left and right by the allowed window
            for column in range(row - window_size, row + 1 + window_size):
                if column < 1 or column >= m:
                    continue

                # find smallest entry in the warping matrix, either above, to the left, or diagonally left and up
                above = warp_matrix[row - 1][column]
                left = warp_matrix[row][column - 1]
                diag = warp_matrix[row - 1][column - 1]

                # add the pairwise distance for [row][column] to the minimum of the three possible potential cells
                warp_matrix[row][column] = pairwise_distances[row][column] + min([above, left, diag])

                # check for evidence that cutoff has been beaten on this row (if using)
                if cutoff is not None and warp_matrix[row][column] < cutoff:
                    cutoff_beaten = True

            # if using a cutoff, at least one calculated value on this row MUST be less than the cutoff otherwise the
            # final distance is guaranteed not to be less. Therefore, if the cutoff has not been beaten, early-abandon
            if cutoff is not None and cutoff_beaten is False:
                return float("inf")

        return warp_matrix[n - 1][m - 1]

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return dtw_single_channel(first, second, **kwargs)

    dist = 0
    for dim in range(0, len(first)):
        dist += dtw_single_channel(first[dim], second[dim], **kwargs)

    return dist


def derivative_dtw_distance(first, second, **kwargs):

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return dtw_distance(np.diff(first), np.diff(second), **kwargs)

    dist = 0
    for dim in range(0, len(first)):
        dist += dtw_distance([first[dim].diff()[1:]], [second[dim].diff()[1:]], **kwargs)
    return dist


def weighted_dtw_distance(first, second, **kwargs):

    def wdtw_single_channel(first, second, **kwargs):
        try:
            g = kwargs["g"]
        except:
            g = 0.0

        m = len(first);
        n = len(second);

        weight_vector = [1/(1+np.exp(-g*(i-m/2))) for i in range(0,m)]

        dist = lambda x1, x2: ((x1 - x2) ** 2)
        pairwise_distances = np.asarray([[dist(x1, x2) for x2 in second] for x1 in first])

        # initialise edges of the warping matrix
        distances = np.full([n, m], np.inf)
        distances[0][0] = weight_vector[0]*pairwise_distances[0][0]

        # top row
        for i in range(1,n):
            distances[0][i] = distances[0][i - 1] + weight_vector[i] * pairwise_distances[0][i]

        # first column
        for i in range(1, m):
            distances[i][0] = distances[i - 1][0] + weight_vector[i] * pairwise_distances[i][0]

        # warp rest
        for i in range(1,m):
            for j in range(1,n):
                min_dist = np.min([distances[i][j - 1], distances[i - 1][j], distances[i - 1][j - 1]])
                # print(min_dist)
                distances[i][j] = min_dist+weight_vector[np.abs(i-j)] * pairwise_distances[i][j]
        return distances[m-1][n-1]

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return wdtw_single_channel(first, second, **kwargs)
    dist = 0
    for dim in range(0, len(first)):
        dist += wdtw_single_channel(first[dim], second[dim], **kwargs)

    return dist


def weighted_derivative_dtw_distance(first, second, **kwargs):

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return weighted_dtw_distance(np.diff(first), np.diff(second), **kwargs)

    dist = 0
    for dim in range(0, len(first)):
        dist += weighted_dtw_distance([first[dim].diff()[1:]], [second[dim].diff()[1:]], **kwargs)
    return dist


def lcss_distance(first, second, **kwargs):

    def lcss_single_channel(first, second, **kwargs) -> float:
        try:
            delta = kwargs["delta"]
        except KeyError:
            delta = 3

        try:
            epsilon = kwargs["epsilon"]
        except KeyError:
            epsilon = 1

        m = len(first)
        n = len(second)

        lcss = np.zeros([m + 1, n + 1])

        for i in range(0, m):
            # for (int j = i-delta; j <= i+delta; j++){
            for j in range(i - delta, i + delta + 1):
                if j < 0:
                    j = -1
                elif j >= n:
                    j = i + delta
                elif second[j] + epsilon >= first[i] and second[j] - epsilon <= first[i]:
                    lcss[i + 1][j + 1] = lcss[i][j] + 1
                elif lcss[i][j + 1] > lcss[i + 1][j]:
                    lcss[i + 1][j + 1] = lcss[i][j + 1]
                else:
                    lcss[i + 1][j + 1] = lcss[i + 1][j]

        max_val = -1;
        for i in range(1, len(lcss[len(lcss) - 1])):
            if lcss[len(lcss) - 1][i] > max_val:
                max_val = lcss[len(lcss) - 1][i];

        return 1 - (max_val / m);

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return lcss_single_channel(first, second, **kwargs)

    dist = 0
    for dim in range(0,len(first)):
        dist += lcss_single_channel(first[dim],second[dim],**kwargs)
    return dist


def msm_distance(first, second, **kwargs):

    def msm_single_channel(first, second, **kwargs) -> float:
        try:
            c = kwargs["c"]
        except KeyError:
            c = 0.1

        m = len(first)
        n = len(second)

        cost = np.zeros([m, n])

        def calc_cost(new_point, x, y):
            dist = 0

            if ((x <= new_point) and (new_point <= y)) or ((y <= new_point) and (new_point <= x)):
                return c
            else:
                return c + min(np.abs(new_point - x), np.abs(new_point - y))

        # Initialization
        cost[0][0] = np.abs(first[0] - second[0])
        for i in range(1, m):
            cost[i][0] = cost[i - 1][0] + calc_cost(first[i], first[i - 1], second[0])

        for i in range(1, n):
            cost[0][i] = cost[0][i - 1] + calc_cost(second[i], first[0], second[i - 1])

        # Main Loop
        for i in range(1, m):
            for j in range(1, n):
                d1 = cost[i - 1][j - 1] + np.abs(first[i] - second[j])
                d2 = cost[i - 1][j] + calc_cost(first[i], first[i - 1], second[j])
                d3 = cost[i][j - 1] + calc_cost(second[j], first[i], second[j - 1])
                cost[i][j] = min(d1, d2, d3)

        return cost[m - 1][n - 1];

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return msm_single_channel(first, second, **kwargs)

    dist = 0
    for dim in range(0, len(first)):
        dist += msm_single_channel(first[dim], second[dim], **kwargs)
    return dist


def erp_distance(first, second, **kwargs):

    def erp_single_channel(first, second, **kwargs):
        """
        Adapted from:
            This file is part of ELKI:
            Environment for Developing KDD-Applications Supported by Index-Structures

            Copyright (C) 2011
            Ludwig-Maximilians-UniversitÃ¤t MÃ¼nchen
            Lehr- und Forschungseinheit fÃ¼r Datenbanksysteme
            ELKI Development Team

            This program is free software: you can redistribute it and/or modify
            it under the terms of the GNU Affero General Public License as published by
            the Free Software Foundation, either version 3 of the License, or
            (at your option) any later version.

            This program is distributed in the hope that it will be useful,
            but WITHOUT ANY WARRANTY; without even the implied warranty of
            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
            GNU Affero General Public License for more details.

            You should have received a copy of the GNU Affero General Public License
            along with this program.  If not, see <http://www.gnu.org/licenses/>.
        """

        try:
            band_size = kwargs["band_size"]
        except KeyError:
            band_size = 5
        try:
            g = kwargs["g"]
        except KeyError:
            g = 0.5

        m = len(first)
        n = len(second)
        band = np.ceil(band_size * m)
        curr = np.empty(m)
        prev = np.empty(m)

        for i in range(0, m):
            temp = prev
            prev = curr
            curr = temp
            l = i - (band + 1)

            if l < 0:
                l = 0

            r = i + (band + 1);
            if r > m - 1:
                r = (m - 1)

            for j in range(l, r + 1):
                if np.abs(i - j) <= band:
                    val1 = first[i]
                    val2 = g
                    diff = (val1 - val2)
                    d1 = np.sqrt(diff * diff)

                    val1 = g
                    val2 = second[j]
                    diff = (val1 - val2)
                    d2 = np.sqrt(diff * diff)

                    val1 = first[i]
                    val2 = second[j]
                    diff = (val1 - val2)
                    d12 = np.sqrt(diff * diff)

                    dist1 = d1 * d1
                    dist2 = d2 * d2
                    dist12 = d12 * d12

                    if i + j != 0:
                        if i == 0 or (j != 0 and (((prev[j - 1] + dist12) > (curr[j - 1] + dist2)) and ((curr[j - 1] + dist2) < (prev[j] + dist1)))):
                            # del
                            cost = curr[j - 1] + dist2
                        elif (j == 0) or ((i != 0) and (((prev[j - 1] + dist12) > (prev[j] + dist1)) and ((prev[j] + dist1) < (curr[j - 1] + dist2)))):
                            # ins
                            cost = prev[j] + dist1;
                        else:
                            # match
                            cost = prev[j - 1] + dist12
                    else:
                        cost = 0

                    curr[j] = cost

        return np.sqrt(curr[m - 1]);

    if isinstance(first, np.ndarray) and isinstance(first[0], float) is True:
        return erp_single_channel(first, second, **kwargs)

    dist = 0
    for dim in range(0, len(first)):
        dist += erp_single_channel(first[dim], second[dim], **kwargs)
    return dist



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-

__author__ = ["Markus Löning"]
__all__ = [
    "ForecastingHorizon",
    "load_lynx",
    "load_longley",
    "load_airline",
    "load_shampoo_sales",
    "CutoffSplitter",
    "ForecastingGridSearchCV",
    "SlidingWindowSplitter",
    "SingleWindowSplitter",
    "temporal_train_test_split",
    "NaiveForecaster",
    "ExponentialSmoothing",
    "ThetaForecaster",
    "AutoARIMA",
    "PolynomialTrendForecaster",
    "TransformedTargetForecaster",
    "Deseasonalizer",
    "ReducedRegressionForecaster",
    "ReducedTimeSeriesRegressionForecaster",
    "EnsembleForecaster",
    "Detrender",
    "sMAPE",
    "MASE",
    "smape_loss",
    "mase_loss",
    "pd",
    "np",
    "plot_series",
    "NormalHedgeEnsemble",
    "NNLSEnsemble",
    "OnlineEnsembleForecaster",
]

import numpy as np
import pandas as pd

from sktime.datasets import load_airline
from sktime.datasets import load_longley
from sktime.datasets import load_lynx
from sktime.datasets import load_shampoo_sales
from sktime.forecasting.arima import AutoARIMA
from sktime.forecasting.base import ForecastingHorizon
from sktime.forecasting.compose import EnsembleForecaster
from sktime.forecasting.compose import ReducedRegressionForecaster
from sktime.forecasting.compose import ReducedTimeSeriesRegressionForecaster
from sktime.forecasting.compose import TransformedTargetForecaster
from sktime.forecasting.exp_smoothing import ExponentialSmoothing
from sktime.forecasting.model_selection import CutoffSplitter
from sktime.forecasting.model_selection import ForecastingGridSearchCV
from sktime.forecasting.model_selection import SingleWindowSplitter
from sktime.forecasting.model_selection import SlidingWindowSplitter
from sktime.forecasting.model_selection import temporal_train_test_split
from sktime.forecasting.naive import NaiveForecaster
from sktime.forecasting.theta import ThetaForecaster
from sktime.forecasting.trend import PolynomialTrendForecaster
from sktime.performance_metrics.forecasting import MASE
from sktime.performance_metrics.forecasting import mase_loss
from sktime.performance_metrics.forecasting import sMAPE
from sktime.performance_metrics.forecasting import smape_loss
from sktime.transformations.series.detrend import Deseasonalizer
from sktime.transformations.series.detrend import Detrender
from sktime.utils.plotting import plot_series
from sktime.forecasting.online_learning._prediction_weighted_ensembler import (
    NormalHedgeEnsemble,
    NNLSEnsemble,
)
from sktime.forecasting.online_learning._online_ensemble import (
    OnlineEnsembleForecaster,
)



# -*- coding: utf-8 -*-
import numpy as np
from sklearn.model_selection import train_test_split

from sktime.classification.interval_based import TimeSeriesForest
from sktime.datasets import load_gunpoint

# expected y_proba
expected = np.array(
    [
        [1.0, 0.0],
        [1.0, 0.0],
        [0.9, 0.1],
        [1.0, 0.0],
        [0.0, 1.0],
        [0.95, 0.05],
        [0.0, 1.0],
        [0.9, 0.1],
        [1.0, 0.0],
        [0.15, 0.85],
        [0.8, 0.2],
        [1.0, 0.0],
        [1.0, 0.0],
        [0.25, 0.75],
        [1.0, 0.0],
        [0.95, 0.05],
        [1.0, 0.0],
        [1.0, 0.0],
        [0.0, 1.0],
        [0.0, 1.0],
    ]
)


def test_y_proba_on_gunpoint():
    X, y = load_gunpoint(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.1, random_state=42
    )
    estimator = TimeSeriesForest(random_state=42, n_estimators=20)
    estimator.fit(X_train, y_train)
    actual = estimator.predict_proba(X_test)
    np.testing.assert_array_equal(actual, expected)



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import numbers
import math
from sktime.utils.data_processing import from_nested_to_2d_array
from sktime.utils.validation.panel import check_X
from sktime.transformations.base import _PanelToPanelTransformer

"""
The HOG1D Transformer proposed by:

@article{zhao2015classifying,
  title={Classifying time series using local descriptors with hybrid sampling},
  author={Zhao, Jiaping and Itti, Laurent},
  journal={IEEE Transactions on Knowledge and Data Engineering},
  volume={28},
  number={3},
  pages={623--637},
  year={2015},
  publisher={IEEE}
}
"""


class HOG1DTransformer(_PanelToPanelTransformer):

    """
    This class is to calculate the HOG1D transform of a
    dataframe of time series data. Works by splitting
    the time series num_intervals times, and calculate
    a histogram of gradients within each interval.

    Parameters
    ----------
        num_intervals   : int, length of interval.
        num_bins        : int, num bins in the histogram.
        scaling_factor  : float, a constant that is multiplied
                          to modify the distribution.
    """

    def __init__(self, num_intervals=2, num_bins=8, scaling_factor=0.1):
        self.num_intervals = num_intervals
        self.num_bins = num_bins
        self.scaling_factor = scaling_factor
        super(HOG1DTransformer, self).__init__()

    def transform(self, X, y=None):
        """
        Function to transform a data frame of time series data.

        Parameters
        ----------
        X : a pandas dataframe of shape = [n_samples, num_dims]
            The training input samples.

        Returns
        -------
        dims: a pandas data frame of shape = [n_samples, num_dims]
        """

        # Check the data
        self.check_is_fitted()
        X = check_X(X, enforce_univariate=False, coerce_to_pandas=True)

        # Get information about the dataframe
        num_insts = X.shape[0]
        col_names = X.columns
        num_atts = len(X.iloc[0, 0])

        # Check the parameters are appropriate
        self._check_parameters(num_atts)

        df = pd.DataFrame()

        for x in col_names:
            # Convert one of the columns in the dataframe to a numpy array
            arr = from_nested_to_2d_array(pd.DataFrame(X[x]), return_numpy=True)

            # Get the HOG1Ds of each time series
            transformedData = []
            for y in range(num_insts):
                inst = self._calculate_hog1ds(arr[y])
                transformedData.append(inst)

            # Convert to numpy array
            transformedData = np.asarray(transformedData)

            # Add it to the dataframe
            colToAdd = []
            for i in range(len(transformedData)):
                inst = transformedData[i]
                colToAdd.append(pd.Series(inst))

            df[x] = colToAdd

        return df

    def _calculate_hog1ds(self, X):
        """
        Function to calculate the HOG1Ds given a time series.

        Parameters
        ----------
        X : a numpy array of shape = [time_series_length]

        Returns
        -------
        HOG1Ds : a numpy array of shape = [num_intervals*num_bins].
                 It contains the histogram of each gradient within
                 each interval.
        """
        # Firstly, split the time series into approx equal
        # length intervals
        splitTimeSeries = self._split_time_series(X)
        HOG1Ds = []

        for x in range(len(splitTimeSeries)):
            HOG1Ds.extend(self._get_hog1d(splitTimeSeries[x]))

        return HOG1Ds

    def _get_hog1d(self, X):
        """
        Function to get the HOG1D given a portion of a time series.

        X : a numpy array of shape = [interval_size]

        Returns
        -------
        histogram : a numpy array of shape = [num_bins].
        """
        # First step is to pad the portion on both ends once.
        gradients = [0.0] * (len(X))
        X = np.pad(X, 1, mode="edge")
        histogram = [0.0] * self.num_bins

        # Calculate the gradients of each element
        for i in range(1, len(X) - 1):
            gradients[(i - 1)] = self.scaling_factor * 0.5 * (X[(i + 1)] - X[(i - 1)])

        # Calculate the orientations
        orients = [math.degrees(math.atan(x)) for x in gradients]

        # Calculate the boundaries of the histogram
        hisBoundaries = [
            -90 + (180 / self.num_bins) + ((180 / self.num_bins) * x)
            for x in range(self.num_bins)
        ]

        # Construct the histogram
        for x in range(len(orients)):
            orientToAdd = orients[x]
            for y in range(len(hisBoundaries)):
                if orientToAdd <= hisBoundaries[y]:
                    histogram[y] += 1.0
                    break

        return histogram

    def _split_time_series(self, X):
        """
        Function to split a time series into approximately equal intervals.

        Adopted from = https://stackoverflow.com/questions/2130016/splitting
                       -a-list-into-n-parts-of-approximately-equal-length

        Parameters
        ----------
        X : a numpy array corresponding to the time series being split
            into approx equal length intervals of shape
            [num_intervals,interval_length].
        """
        avg = len(X) / float(self.num_intervals)
        output = []
        beginning = 0.0

        while beginning < len(X):
            output.append(X[int(beginning) : int(beginning + avg)])
            beginning += avg

        return output

    def _check_parameters(self, num_atts):
        """
        Function for checking the values of parameters inserted into HOG1D.

        Throws
        ------
        ValueError or TypeError if a parameters input is invalid.
        """
        if isinstance(self.num_intervals, int):
            if self.num_intervals <= 0:
                raise ValueError(
                    "num_intervals must have \
                                  the value of at least 1"
                )
            if self.num_intervals > num_atts:
                raise ValueError(
                    "num_intervals cannot be higher \
                                  than subsequence_length"
                )
        else:
            raise TypeError(
                "num_intervals must be an 'int'. \
                            Found '"
                + type(self.num_intervals).__name__
                + "' instead."
            )

        if isinstance(self.num_bins, int):
            if self.num_bins <= 0:
                raise ValueError(
                    "num_bins must have the value of \
                                  at least 1"
                )
        else:
            raise TypeError(
                "num_bins must be an 'int'. Found '"
                + type(self.num_bins).__name__
                + "' \
                            instead."
            )

        if not isinstance(self.scaling_factor, numbers.Number):
            raise TypeError(
                "scaling_factor must be a 'number'. \
                            Found '"
                + type(self.scaling_factor).__name__
                + "' instead."
            )



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Markus Löning"]
__all__ = ["_HeterogenousMetaEstimator"]

from abc import ABCMeta

from sktime.base import BaseEstimator


class _HeterogenousMetaEstimator(BaseEstimator, metaclass=ABCMeta):
    """Handles parameter management for estimtators composed of named
    estimators.

    from sklearn utils.metaestimator.py
    """

    def get_params(self, deep=True):
        raise NotImplementedError("abstract method")

    def set_params(self, **params):
        raise NotImplementedError("abstract method")

    def _get_params(self, attr, deep=True):
        out = super().get_params(deep=deep)
        if not deep:
            return out
        estimators = getattr(self, attr)
        out.update(estimators)
        for name, estimator in estimators:
            if hasattr(estimator, "get_params"):
                for key, value in estimator.get_params(deep=True).items():
                    out["%s__%s" % (name, key)] = value
        return out

    def _set_params(self, attr, **params):
        # Ensure strict ordering of parameter setting:
        # 1. All steps
        if attr in params:
            setattr(self, attr, params.pop(attr))
        # 2. Step replacement
        items = getattr(self, attr)
        names = []
        if items:
            names, _ = zip(*items)
        for name in list(params.keys()):
            if "__" not in name and name in names:
                self._replace_estimator(attr, name, params.pop(name))
        # 3. Step parameters and other initialisation arguments
        super().set_params(**params)
        return self

    def _replace_estimator(self, attr, name, new_val):
        # assumes `name` is a valid estimator name
        new_estimators = list(getattr(self, attr))
        for i, (estimator_name, _) in enumerate(new_estimators):
            if estimator_name == name:
                new_estimators[i] = (name, new_val)
                break
        setattr(self, attr, new_estimators)

    def _check_names(self, names):
        if len(set(names)) != len(names):
            raise ValueError("Names provided are not unique: {0!r}".format(list(names)))
        invalid_names = set(names).intersection(self.get_params(deep=False))
        if invalid_names:
            raise ValueError(
                "Estimator names conflict with constructor "
                "arguments: {0!r}".format(sorted(invalid_names))
            )
        invalid_names = [name for name in names if "__" in name]
        if invalid_names:
            raise ValueError(
                "Estimator names must not contain __: got "
                "{0!r}".format(invalid_names)
            )



#!/usr/bin/env python3 -u
# coding: utf-8
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Markus Löning"]

__all__ = [
    "EnsembleForecaster",
    "TransformedTargetForecaster",
    "DirectRegressionForecaster",
    "DirectTimeSeriesRegressionForecaster",
    "RecursiveRegressionForecaster",
    "RecursiveTimeSeriesRegressionForecaster",
    "ReducedRegressionForecaster",
    "ReducedTimeSeriesRegressionForecaster",
    "StackingForecaster",
]

from sktime.forecasting.compose._ensemble import EnsembleForecaster
from sktime.forecasting.compose._pipeline import TransformedTargetForecaster
from sktime.forecasting.compose._reduce import DirectRegressionForecaster
from sktime.forecasting.compose._reduce import \
    DirectTimeSeriesRegressionForecaster
from sktime.forecasting.compose._reduce import RecursiveRegressionForecaster
from sktime.forecasting.compose._reduce import \
    RecursiveTimeSeriesRegressionForecaster
from sktime.forecasting.compose._reduce import ReducedRegressionForecaster
from sktime.forecasting.compose._reduce import \
    ReducedTimeSeriesRegressionForecaster
from sktime.forecasting.compose._stack import StackingForecaster



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Markus Löning"]
__all__ = ["PresplitFilesCV", "SingleSplit"]

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split


class PresplitFilesCV:
    """
    Cross-validation iterator over split predefined in files.

    This class is useful in orchestration where the train and test set
    is provided in separate files.
    """

    def __init__(self, cv=None):
        self.cv = cv

    def split(self, data, y=None, groups=None):
        """
        Split the data according to the train/test index.

        Parameters
        ----------
        data : pandas.DataFrame

        Yields
        ------
        train : ndarray
            Train indicies
        test : ndarray
            Test indices
        """
        # check input
        if not isinstance(data, pd.DataFrame):
            raise ValueError(f"Data must be pandas DataFrame, but found {type(data)}")
        if not np.all(data.index.unique().isin(["train", "test"])):
            raise ValueError(
                "Train-test split not properly defined in "
                "index of passed pandas DataFrame"
            )

        # this is a bit of a hack, PresplitFilesCV would need to talk to the
        # data loader during orchestration,
        # but since this is not possible with sklearn's interface, we load
        # both files separately, combined them
        # keep the training and test split in the index of the combined
        # dataframe, and split them here again
        n_instances = data.shape[0]
        idx = np.arange(n_instances)
        train = idx[data.index == "train"]
        test = idx[data.index == "test"]
        yield train, test

        # if additionally a cv iterator is provided, yield the predefined
        # split first, then reshuffle and apply cv,
        # note that test sets may overlap with the presplit file test set
        if self.cv is not None:
            for train, test in self.cv.split(idx, y=y):
                yield train, test

    def get_n_splits(self):
        n_splits = 1 if self.cv is None else 1 + self.cv.get_n_splits()
        return n_splits


class SingleSplit:
    """
    Helper class for orchestration that uses a single split for training and
    testing. Wrapper for sklearn.model_selection.train_test_split

    Parameters
    ----------
    *arrays : sequence of indexables with same length / shape[0]
        Allowed inputs are lists, numpy arrays, scipy-sparse
        matrices or pandas dataframes.
    test_size : float, int or None, optional (default=0.25)
        If float, should be between 0.0 and 1.0 and represent the proportion
        of the dataset to include in the test split. If int, represents the
        absolute number of test samples. If None, the value is set to the
        complement of the train size. By default, the value is set to 0.25.
        The default will change in version 0.21. It will remain 0.25 only
        if ``train_size`` is unspecified, otherwise it will complement
        the specified ``train_size``.
    train_size : float, int, or None, (default=None)
        If float, should be between 0.0 and 1.0 and represent the
        proportion of the dataset to include in the train split. If
        int, represents the absolute number of train samples. If None,
        the value is automatically set to the complement of the test size.
    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    shuffle : boolean, optional (default=True)
        Whether or not to shuffle the data before splitting. If shuffle=False
        then stratify must be None.
    stratify : array-like or None (default=None)
        If not None, data is split in a stratified fashion, using this as
        the class labels.
    """

    def __init__(
        self,
        test_size=0.25,
        train_size=None,
        random_state=None,
        shuffle=True,
        stratify=None,
    ):
        self._test_size = test_size
        self._train_size = train_size
        self._random_state = random_state
        self._shuffle = shuffle
        self._stratify = stratify

    def split(self, data, y=None, groups=None):
        """
        Parameters
        ---------
        data : pandas dataframe
            data used for cross validation

        Returns
        -------
        tuple
            (train, test) indexes
        """
        if not isinstance(data, pd.DataFrame):
            raise ValueError("Data must be provided as a pandas DataFrame")
        n_instances = data.shape[0]
        idx = np.arange(n_instances)

        yield train_test_split(
            idx,
            test_size=self._test_size,
            train_size=self._train_size,
            random_state=self._random_state,
            shuffle=self._shuffle,
            stratify=self._stratify,
        )

    @staticmethod
    def get_n_splits():
        return 1



# -*- coding: utf-8 -*-
from sklearn.model_selection import GridSearchCV
from sktime.base import BaseEstimator

from sktime.classification.distance_based._time_series_neighbors import (
    KNeighborsTimeSeriesClassifier,
)
from sktime.datasets import load_italy_power_demand
from sklearn.pipeline import Pipeline
from sktime.transformations.panel.dictionary_based import SAX


class BagOfPatterns(BaseEstimator):
    __author__ = "Matthew Middlehurst"
    """ Bag of Patterns classifier
    """

    def bop_pipeline(X, y):
        steps = [
            ("transform", SAX(remove_repeat_words=True)),
            (
                "clf",
                KNeighborsTimeSeriesClassifier(
                    n_neighbors=1, metric=euclidean_distance
                ),
            ),
        ]
        pipeline = Pipeline(steps)

        series_length = X.iloc[0, 0].shape[0]
        max_window_searches = series_length / 4
        win_inc = int((series_length - 10) / max_window_searches)
        if win_inc < 1:
            win_inc = 1
        window_sizes = [win_size for win_size in range(10, series_length + 1, win_inc)]

        cv_params = {
            "transform__word_length": [8, 10, 12, 14, 16],
            "transform__alphabet_size": [2, 3, 4],
            "transform__window_size": window_sizes,
        }
        model = GridSearchCV(pipeline, cv_params, cv=5)
        model.fit(X, y)
        return model


def euclidean_distance(first, second, best_dist=sys.float_info.max):
    dist = 0

    if isinstance(first, dict):
        words = set(list(first) + list(second))
        for word in words:
            val_a = first.get(word, 0)
            val_b = second.get(word, 0)
            dist += (val_a - val_b) * (val_a - val_b)

            if dist > best_dist:
                return sys.float_info.max
    else:
        dist = np.sum(
            [(first[n] - second[n]) * (first[n] - second[n]) for n in range(len(first))]
        )

    return dist


if __name__ == "__main__":
    X_train, y_train = load_italy_power_demand(split="TRAIN", return_X_y=True)
    X_test, y_test = load_italy_power_demand(split="TEST", return_X_y=True)

    model = bop_pipeline(X_train, y_train)
    model.predict(X_test)
